import os
import shutil
import tempfile
from subprocess import run

import numpy as np
from pkg_resources import resource_filename

from tinamit.config import _
from tinamit.envolt.bf import ModeloBloques
from ._arch_egr import leer_arch_egr
from ._arch_ingr import leer_info_dic_paráms, escribir_desde_dic_paráms
from ._vars import VariablesSAHYSMOD, VarBloqSAHYSMOD


class ModeloSAHYSMOD(ModeloBloques):
    """
    Envoltura para modelos SAHYSMOD.
    """

    idioma_orig = 'en'  # La lengua de los nombres y descripción de los variables (y NO la del código aquí)

    def __init__(símismo, archivo, nombre='SAHYSMOD'):

        símismo.archivo = archivo

        símismo.direc_trabajo = ''

        # Directorio vacío para guardar datos de ingresos después
        símismo.dic_ingr = {}

        # Buscar la ubicación del modelo SAHYSMOD.
        símismo.exe_SAHYSMOD = símismo.obt_conf(
            'exe',
            cond=os.path.isfile,
            mnsj_err=_(
                '\nDebes especificar la ubicación del ejecutable SAHYSMOD, p. ej.'
                '\n\tModeloSAHYSMOD.estab_conf("exe", "C:\\Camino\\hacia\\mi\\SAHYSMODConsole.exe")'
                '\npara poder hacer simulaciones con modelos SAHYSMOD.'
                '\nSi no instalaste SAHYSMOD, lo puedes conseguir para Linux, Mac o Windows de '
                'https://github.com/julienmalard/sahysmod-sourcecode.'
            )
        )

        # Leer el fuente de ingreso
        símismo.dic_ingr = leer_info_dic_paráms(archivo_fnt=símismo.archivo)

        variables = VariablesSAHYSMOD(inic=símismo.dic_ingr)

        # Inicializar la clase pariente.
        super().__init__(variables=variables, nombre=nombre)

        # Establecer los variables climáticos.
        símismo.conectar_var_clima(var='Pp - Rainfall', var_clima='بارش', combin='total', conv=0.001)

    def iniciar_modelo(símismo, corrida):

        # Crear un diccionario de trabajo específico a esta corrida.
        símismo.direc_trabajo = tempfile.mkdtemp('_' + str(corrida))
        super().iniciar_modelo(corrida)

    def avanzar_modelo(símismo, n_ciclos):
        arch_egreso = os.path.join(símismo.direc_trabajo, 'SAHYSMOD.out')
        arch_ingreso = os.path.join(símismo.direc_trabajo, 'SAHYSMOD.inp')

        símismo._verificar_estado_vars()
        símismo._escribir_archivo_ingr(n_ciclos=n_ciclos, archivo=arch_ingreso)

        # Limpiar archivos de egresos que podrían estar allí
        if os.path.isfile(arch_egreso):
            os.remove(arch_egreso)

        # Correr la comanda desde la línea de comanda
        comanda = '"{SAHYSMOD}" "{ingreso}" "{egreso}"'.format(
            SAHYSMOD=símismo.exe_SAHYSMOD, ingreso=arch_ingreso, egreso=arch_egreso
        )
        run(comanda, cwd=símismo.direc_trabajo)

        # Verificar que SAHYSMOD generó egresos.
        if not os.path.isfile(arch_egreso):
            with open(os.path.join(símismo.direc_trabajo, 'error.lst')) as d:
                mnsj_sahysmod = d.readlines()

            raise FileNotFoundError(_(
                '\nEl modelo SAHYSMOD no generó egreso. Esto probablemente quiere decir que tuvo problema interno.'
                '\n¡Diviértete! :)'
                '\nMensajes de error de SAHYSMOD:'
                '\n{}').format(mnsj_sahysmod))

        símismo.leer_egr_modelo(n_ciclos=n_ciclos)

    def cerrar(símismo):
        shutil.rmtree(símismo.direc_trabajo)

    def leer_egr_modelo(símismo, n_ciclos):
        archivo = os.path.join(símismo.direc_trabajo, 'SAHYSMOD.out')

        dic_egr = leer_arch_egr(
            archivo=archivo, años=n_ciclos
        )

        # Convertir códigos de variables a nombres de variables
        for c, v in dic_egr.items():
            símismo.variables.cód_a_var[c].poner_val(v[0])  # v[0] para quitar dimensión de año

    @classmethod
    def instalado(cls):
        return cls.obt_conf('exe') is not None

    def paralelizable(símismo):
        return True

    def unidad_tiempo(símismo):
        return 'mes'

    @classmethod
    def prb_egreso(cls):
        arch = resource_filename(__name__, 'rcrs/prb_egresos.out')

        return arch, leer_arch_egr

    @classmethod
    def prb_ingreso(cls):
        arch = resource_filename(__name__, 'rcrs/prb_ingresos.inp')

        def f(a):
            return VariablesSAHYSMOD(leer_info_dic_paráms(a))

        return arch, f

    @classmethod
    def prb_simul(cls):
        return resource_filename(__name__, 'rcrs/prb_ingresos.inp')

    def _escribir_archivo_ingr(símismo, n_ciclos, archivo):

        # Establecer el número de años de simulación
        símismo.dic_ingr['NY'] = n_ciclos

        # Copiar datos desde el diccionario de ingresos
        for var in símismo.variables.egresos():
            llave = var.cód.replace('#', '').upper()
            if llave in símismo.dic_ingr:
                m = símismo.dic_ingr[llave]

                if isinstance(var, VarBloqSAHYSMOD):
                    val = var.obt_vals_paso()
                    if m.shape == val.shape:
                        m[:] = val
                    else:
                        m[:] = val[-1]
                else:
                    val = var.obt_val()
                    m[:] = val
                m[np.isnan(m)] = -1

        # Y finalmente, escribir el fuente de valores de ingreso
        escribir_desde_dic_paráms(dic_paráms=símismo.dic_ingr, archivo_obj=archivo)

    def _verificar_estado_vars(símismo):
        
        # Aquí tenemos que verificar el estado interno de SAHYSMOD porque éste, siendo SAHYSMOD, da mensajes de error
        # con el mínimo de información posible.

        a = símismo.variables['Area A - Seasonal fraction area crop A']
        b = símismo.variables['Area B - Seasonal fraction area crop B']
        fsa = símismo.variables['FsA - Water storage efficiency crop A']
        fsb = símismo.variables['FsB - Water storage efficiency crop B']

        if np.any(np.logical_and(fsa == -1, a > 0)):
            raise ValueError(_('Los valores de FsA no pueden faltar en polígonos que tienen superficie con cultivo A.'))
        
        if np.any(np.logical_and(fsb == -1, b > 0)):
            raise ValueError(_('Los valores de FsB no pueden faltar en polígonos que tienen superficie con cultivo B.'))



from lark import Lark
from tinamit.config import _

from ._vstr import VstrSacarVars, VstrCoefDe, VstrExtrArgs, VstrContarVar, TrnsfSubstNmbrs
from ._vstr_transf import VstrATx, VstrAPyMC3, VstrAPy

try:
    import pymc3 as pm
except ImportError:  # pragma: sin cobertura
    pm = None

from pkg_resources import resource_filename


class Gramática(object):
    _arch = resource_filename('tinamit.calibs.sintx', 'gram_ecs.g')

    def __init__(símismo):
        with open(símismo._arch, encoding='UTF-8') as d:
            símismo.gram = Lark(d, parser='lalr', start='ec')


anlzdr = Gramática()


class Ecuación(object):
    """
    Un objeto para manejar ecuaciones dinámicas.
    """

    def __init__(símismo, ec, nombre=None, otras_ecs=None, nombres_equiv=None, dialecto='tinamït'):
        """
        Inicializa un objeto de ecuación.

        Parameters
        ----------
        ec: str
            La ecuación en formato texto.
        nombre: str
            El nombre del variable, si no está especificado en la ecuación sí misma.
        otras_ecs: dict[str, str]
            Un diccionario de ecuaciones para substituir para variables en la ecuación principal.
        nombres_equiv: dict[str, str]
            Un diccionario de cambios de nombres de variables para efectuar.
        """

        # Formatear parámetros vacíos
        if otras_ecs is None:
            otras_ecs = {}

        if nombres_equiv is None:
            nombres_equiv = {}

        # Si es una ecuación de variable o de subscripto
        símismo.tipo = 'var'

        símismo.dialecto = dialecto

        # Analizar la ecuación
        árbol = anlzdr.gram.parse(ec)

        # Aplicar subsituciones de ecuaciones de variables
        subst = {
            v_otr: Ecuación(
                ec_otr, otras_ecs={v: otr for v, otr in otras_ecs.items() if v != v_otr}, dialecto=dialecto
            ).árbol for v_otr, ec_otr in otras_ecs.items()
        }
        TrnsfSubstNmbrs(subst).transform(árbol)

        # Aplicar substituciones de nombres
        TrnsfSubstNmbrs(nombres_equiv).transform(árbol)

        # Analizar los resultados
        if árbol.data == 'asign_var':
            # Si la ecuación incluía el nombre del variable y...
            try:
                # Intentar para declaración de variables
                símismo.nombre = str(árbol.children[0].children[0])
                símismo.árbol = árbol.children[1]
            except KeyError:
                # Si no funcionó, debe sser declaración de subsripto
                símismo.nombre = str(árbol.children[0].children[0])
                símismo.árbol = str(árbol.children[1])
                símismo.tipo = 'sub'
        else:
            # Si no había el nombre del variable y, simplemente aplicamos el árbol.
            símismo.árbol = árbol
            símismo.nombre = nombre if nombre not in nombres_equiv else nombres_equiv[nombre]

    def variables(símismo):
        """
        Devuelve una lista de los variable de esta ecuación.

        Returns
        -------
        set:
            Un conjunto con todos los variables presentes en esta ecuación.
        """

        return VstrSacarVars()(símismo.árbol)

    def a_python(símismo, paráms):
        """
        Genera una función dinámica Python basada en la ecuación. La función generada tomará dos argumentos:
            1. p: una lista con valores para parámetros en el orden especificado en `paráms`.
            2. vr: Un diccionario con valores para cada variable presente en la ecuación y no incluido en `paráms`.

        Parameters
        ----------
        paráms: list
            La lista de parámetros.

        Returns
        -------
        Callable:
            La función dinámica Python.
        """

        return VstrAPy(l_prms=paráms, dialecto=símismo.dialecto).convertir(símismo.árbol)

    def coef_de(símismo, var, l_vars=None):
        # para hacer: ¿necesario?
        if var == símismo.nombre:
            return _coef_de_ec(símismo.árbol, l_vars=l_vars)
        else:
            return VstrCoefDe()(símismo.árbol, var, l_vars)

    def gen_mod_bayes(
            símismo, líms_paráms, obs_x, obs_y, aprioris=None, binario=False, nv_jerarquía=None, í_datos=None,
            trsld_sub=False
    ):

        if pm is None:
            return ImportError(_('Hay que instalar PyMC3 para poder utilizar modelos bayesianos.'))

        def _gen_d_vars_pm(tmñ=(), fmt_nmbrs='{}'):
            egr = {}
            norm = {}
            for p, líms in líms_paráms.items():
                nmbr = fmt_nmbrs.format(p)
                if aprioris is None:
                    if líms[0] is líms[1] is None:
                        final = pm.Normal(name=nmbr, mu=0, sd=100 ** 2, shape=tmñ)
                    else:
                        dist_norm = pm.Normal(name='transf' + nmbr, mu=0, sd=1, shape=tmñ)
                        if líms[0] is None:
                            final = pm.Deterministic(nmbr, líms[1] - pm.math.exp(dist_norm))
                        elif líms[1] is None:
                            final = pm.Deterministic(nmbr, líms[0] + pm.math.exp(dist_norm))
                        else:
                            final = pm.Deterministic(
                                nmbr, (pm.math.tanh(dist_norm) / 2 + 0.5) * (líms[1] - líms[0]) + líms[0]
                            )
                        norm[p] = dist_norm
                else:
                    dist, prms = aprioris[p]
                    if (líms[0] is not None or líms[1] is not None) and dist != pm.Uniform:
                        acotada = pm.Bound(dist, lower=líms[0], upper=líms[1])
                        final = acotada(nmbr, shape=tmñ, **prms)
                    else:
                        if dist == pm.Uniform:
                            prms['lower'] = max(prms['lower'], líms[0])
                            prms['upper'] = min(prms['upper'], líms[1])
                        final = dist(nmbr, shape=tmñ, **prms)

                egr[p] = final
            for p, v in egr.items():
                if p not in norm:
                    norm[p] = v
            return {'final': egr, 'norm': norm}

        def _gen_d_vars_pm_jer():
            dists_base = _gen_d_vars_pm(
                tmñ=(1,), fmt_nmbrs='mu_{}_nv_0' if len(nv_jerarquía) else '{}'
            )['norm']

            egr = {}

            for p, líms in líms_paráms.items():
                mu_v = dists_base[p]

                if líms[0] is líms[1] is None:
                    mu_sg = 2
                else:
                    mu_sg = 0.5

                sg_v = pm.Gamma(name='sg_{}_nv_0'.format(p), mu=mu_sg, sd=0.2, shape=(1,))

                for í, nv in enumerate(nv_jerarquía):
                    tmñ_nv = len(nv)
                    últ_niv = í == (len(nv_jerarquía) - 1)
                    í_nv = í+1

                    nmbr_mu = p if últ_niv else 'mu_{}_nv_{}'.format(p, í_nv)
                    nmbr_sg = 'sg_{}_nv_{}'.format(p, í_nv)
                    nmbr_trsld = 'trlsd_{}_nv_{}'.format(p, í_nv)

                    if False and (trsld_sub or últ_niv):
                        trsld = pm.Normal(name=nmbr_trsld, mu=0, sd=1, shape=tmñ_nv)
                        if líms[0] is líms[1] is None:
                            mu_v = pm.Deterministic(nmbr_mu, mu_v[nv] + trsld * sg_v[nv])

                        else:
                            mu_v = pm.Deterministic('transf' + nmbr_mu, mu_v[nv] + trsld * sg_v[nv])
                            if líms[0] is None:
                                final = pm.Deterministic(nmbr_mu, líms[1] - pm.math.exp(mu_v))
                            elif líms[1] is None:
                                final = pm.Deterministic(nmbr_mu, líms[0] + pm.math.exp(mu_v))
                            else:
                                final = pm.Deterministic(
                                    nmbr_mu, (pm.math.tanh(mu_v) / 2 + 0.5) * (líms[1] - líms[0]) + líms[0]
                                )

                            if últ_niv:
                                mu_v = final
                    else:
                        if líms[0] is líms[1] is None:
                            mu_v = pm.Normal(name=nmbr_mu, mu=mu_v[nv], sd=sg_v[nv], shape=tmñ_nv)
                        else:
                            mu_v = pm.Normal(name='transf' + nmbr_mu, mu=mu_v[nv], sd=sg_v[nv], shape=tmñ_nv)
                            if líms[0] is None:
                                final = pm.Deterministic(nmbr_mu, líms[1] - pm.math.exp(mu_v))
                            elif líms[1] is None:
                                final = pm.Deterministic(nmbr_mu, líms[0] + pm.math.exp(mu_v))
                            else:
                                final = pm.Deterministic(
                                    nmbr_mu, (pm.math.tanh(mu_v) / 2 + 0.5) * (líms[1] - líms[0]) + líms[0]
                                )
                            if últ_niv:
                                mu_v = final

                    if not últ_niv:
                        sg_v = pm.Gamma(name=nmbr_sg, mu=mu_sg, sd=0.2, shape=tmñ_nv)

                egr[p] = mu_v

            return egr

        modelo = pm.Model()
        with modelo:
            if nv_jerarquía is None:
                d_vars_pm = _gen_d_vars_pm()['final']
            else:
                d_vars_pm = _gen_d_vars_pm_jer()

            mu = VstrAPyMC3(
                d_vars_pm=d_vars_pm, obs_x=obs_x, í_datos=í_datos, dialecto=símismo.dialecto
            ).convertir(símismo.árbol)

            if binario:
                pm.Bernoulli(name='Y_obs', logit_p=mu, observed=obs_y.values)

            else:
                sigma = pm.HalfNormal(name='sigma', sd=obs_y.values.std())
                # theta = pm.Normal(name='theta', sd=1)
                # beta = pm.HalfNormal(name='beta', sd=1 / obs_y.values.std())
                pm.Normal(name='Y_obs', mu=mu,
                          sd=sigma,  # * (beta*pm.math.abs_(mu) + 1),
                          observed=obs_y.values)

        return modelo

    def sacar_args_func(símismo, func, i=None):
        """
        Devuelve uno o más argumentos de una función presente en el árbol de la Ecuacion, en formato texto.

        Parameters
        ----------
        func: str
            La función cuyos argumentos buscamos.
        i: int
            El número de argumentos que querremos.

        Returns
        -------
        list[str]
            El argumento de la función.

        """

        # Aplicar la función recursiva.
        return VstrExtrArgs(f=func, i=i, dial='tinamït')(símismo.árbol)

    def __str__(símismo):
        return VstrATx('tinamït').convertir(símismo.árbol)


def _coef_de_ec(árbol, l_vars=None):
    if l_vars is None:
        l_vars = []
    if árbol.data == 'prod':
        arg1 = árbol.children[0]
        arg2 = árbol.children[2]
        op = árbol.children[1].children[0]

        v = inv = ec = None
        if op == '*':
            inv = True  # Y queda del otro lado de la ecuación que el parámetro
            if arg1.data == 'var' and v not in l_vars:
                v = arg1.children[0]
                ec = arg2
            elif arg2.data == 'var' and v not in l_vars:
                v = arg2.children[0]
                ec = arg1
        else:
            if arg1.data == 'var' and v not in l_vars:
                v = arg1.children[0]
                ec = arg2
                inv = True
            elif arg2.data == 'var' and v not in l_vars:
                v = arg2.children[0]
                ec = arg1
                inv = False
        if v is not None and VstrContarVar(v)(ec) == 0:
            return v, inv



import csv

import numpy as np
import xarray as xr

from tinamit.config import _
from tinamit.cositas import valid_nombre_arch, guardar_json
from .var import Variable


class ResultadosGrupo(dict):
    """
    Resultados de una simulación por grupo.
    """

    def __init__(símismo, nombre):
        símismo.nombre = nombre
        super().__init__()


class ResultadosSimul(object):
    """
    Resultados de una simulación.
    """

    def __init__(símismo, nombre, t, vars_interés):
        símismo.nombre = nombre
        símismo.t = t
        símismo.res_vars = {str(v): ResultadosVar(v, t) for v in vars_interés}
        # Para hacer: implementar xr.Dataset una vez se coordinen los nombres de dimensiones de variables

    def actualizar(símismo):
        for v in símismo:
            v.actualizar()

    def guardar(símismo, frmt='json', l_vars=None):
        """
        Guarda los resultados en un archivo.

        Parameters
        ----------
        frmt: str
            El formato deseado. Puede ser ``json`` o ``csv``.
        l_vars:
            La lista de variables de interés.
        """

        if l_vars is None:
            l_vars = list(símismo)
        elif isinstance(l_vars, (str, Variable)):
            l_vars = [l_vars]

        if frmt[0] != '.':
            frmt = '.' + frmt
        arch = valid_nombre_arch(símismo.nombre + frmt)

        if frmt == '.json':
            contenido = símismo.a_dic()
            guardar_json(contenido, arch=arch)

        elif frmt == '.csv':

            with open(arch, 'w', encoding='UTF-8', newline='') as a:
                escr = csv.writer(a)

                escr.writerow([_('fecha')] + símismo.t)
                for var in l_vars:
                    vals = símismo[var].values
                    if len(vals.shape) == 1:
                        escr.writerow([var] + vals.tolist())
                    else:
                        for í in range(vals.shape[1]):
                            escr.writerow(['{}[{}]'.format(var, í)] + vals[:, í].tolist())

        else:
            raise ValueError(_('Formato de resultados "{}" no reconocido.').format(frmt))

    def a_dic(símismo):
        """
        Convierte los resultados en diccionario.

        Returns
        -------
        dict
        """
        return {v: r.a_dic() for v, r in símismo.res_vars.items()}

    def variables(símismo):
        return [v.var for v in símismo]

    def poner_vals_t(símismo, valores):
        for vr, vl in valores.items():
            símismo[vr].poner_vals_t(vl)

    def __str__(símismo):
        return símismo.nombre

    def __iter__(símismo):
        for v in símismo.res_vars.values():
            yield v

    def __getitem__(símismo, itema):
        return símismo.res_vars[str(itema)]

    def __eq__(símismo, otro):
        return símismo.res_vars == otro.res_vars


class ResultadosVar(object):
    """
    Los resultados de un variable.
    """
    def __init__(símismo, var, t):
        símismo.var = var
        símismo.t = t

        dims = símismo.var.dims
        matr = np.zeros((len(t), *dims))
        símismo.vals = xr.DataArray(
            matr, coords={_('fecha'): símismo.t.eje()},
            dims=[_('fecha'), *('x_' + str(i) for i in range(len(dims)))]
        )

    def actualizar(símismo):
        símismo.vals[símismo.t.í // símismo.t.guardar_cada] = símismo.var.obt_val()

    def poner_vals_t(símismo, vals):
        símismo.vals[:] = vals.reshape(símismo.vals.shape)

    def a_dic(símismo):
        return símismo.vals.values

    def interpolar(símismo, fechas):
        eje_ant = símismo.vals[_('fecha')]
        nuevas_fechas = fechas.values[~np.isin(fechas.values, eje_ant.values)]
        vals = símismo.vals
        if nuevas_fechas.size:
            vals = vals.reindex(fecha=np.concatenate((eje_ant.values, nuevas_fechas)))
            vals = vals.sortby(_('fecha'))
            vals = vals.interpolate_na(_('fecha'))
        return vals.where(vals[_('fecha')].isin(fechas), drop=True)

    def __str__(símismo):
        return str(símismo.var)

    def __eq__(símismo, otro):
        return símismo.vals.equals(otro.vals)



import random

import numpy as np

from tinamit.envolt.bf import ModeloBF
from tinamit.mod import VariablesMod, Variable


class PruebaBF(ModeloBF):

    def __init__(símismo, dims_bosques=(1,)):
        vars_ = VariablesMod([
            Variable('Lluvia', unid='m3/mes', ingr=False, egr=True, inic=1, líms=(0, None)),
            Variable('Bosques', unid='m2', ingr=True, egr=False, inic=np.full(dims_bosques, 1e6), líms=(0, None)),
        ])
        super().__init__(variables=vars_)

    def incrementar(símismo, rebanada):
        bosques = símismo.variables['Bosques'].obt_val()

        símismo.variables['Lluvia'].poner_val(random.random() * min(1, np.mean(bosques) / 9e5) ** 5 * rebanada.n_pasos)
        super().incrementar(rebanada)

    def unidad_tiempo(símismo):
        return 'meses'

    def paralelizable(símismo):
        return True



from lark import Tree
from lark.lexer import Token
from lark.visitors import Transformer_InPlace, Visitor
from ._vstr_transf import VstrATx


class TrnsfSubstNmbrs(Transformer_InPlace):
    def __init__(símismo, subst):
        símismo.subst = subst
        super().__init__()

    def var(símismo, x):
        try:
            sub = símismo.subst[x[0]]
            if isinstance(sub, str):
                return Tree('var', Token('NOMBRE', sub))
            else:
                return sub
        except KeyError:
            return Tree('var', x)


class VstrSacarVars(Visitor):
    def __init__(símismo):
        símismo.c_vars = set()

    def var(símismo, x):
        símismo.c_vars.add(str(x.children[0]))

    def __call__(símismo, árbol):
        símismo.c_vars.clear()
        símismo.visit(árbol)
        return símismo.c_vars


class VstrExtrArgs(Visitor):
    def __init__(símismo, f, i, dial):
        símismo.f = f
        símismo.i = i
        símismo.dial = dial
        símismo.args_f = None

    def func(símismo, x):
        if x.children[0] == símismo.f:
            símismo.args_f = [
                VstrATx(dialecto=símismo.dial).convertir(e) for e in x.children[1].children[:símismo.i]
            ]

    def __call__(símismo, árbol):
        símismo.args_f = None
        símismo.visit(árbol)
        return símismo.args_f


class VstrCoefDe(Visitor):
    def __init__(símismo):
        símismo.coef = None
        símismo.inv = None
        símismo.vr = None
        símismo.error = False
        símismo.núm = 0
        símismo.l_vars = []

    def _valid_consist(símismo, coef, inv):
        if símismo.coef is None:
            símismo.coef = coef
            símismo.inv = inv
        else:
            if símismo.coef != coef or símismo.inv != inv:
                símismo.error = True
        símismo.núm += 1

    def prod(símismo, x):
        coef = None
        op = x.children[1].children[0]
        args = [x.children[0], x.children[2]]
        for i, a in enumerate(args):
            if a.data == 'neg':
                a = a.children[0]
            if a.data == 'var':
                a = a.children[0]
            else:
                return

            args[i] = a

        if op == '*':
            inv = False

            if args[0] == símismo.vr and args[1] not in símismo.l_vars:
                coef = str(args[1])
            elif args[1] == símismo.vr and args[0] not in símismo.l_vars:
                coef = str(args[0])
        else:
            inv = True
            if args[0] == símismo.vr and args[1] not in símismo.l_vars:
                coef = str(args[1])
            elif args[1] == símismo.vr and args[0] not in símismo.l_vars:
                coef = str(args[0])

        if coef is not None:
            símismo._valid_consist(coef, inv)

    def __call__(símismo, árbol, var, l_vars=None):
        símismo.núm = 0
        símismo.vr = var
        símismo.coef = símismo.inv = None
        símismo.l_vars = [] if l_vars is None else l_vars
        símismo.visit(árbol)
        if símismo.error or símismo.coef is None:
            return None
        elif not (símismo.núm == VstrContarVar(símismo.coef)(árbol) == VstrContarVar(símismo.vr)(árbol)):
            return None
        else:
            return símismo.coef, símismo.inv


class VstrContarVar(Visitor):
    def __init__(símismo, var):
        símismo.vr = var
        símismo.cnt = 0

    def var(símismo, x):
        if x.children[0] == símismo.vr:
            símismo.cnt += 1

    def __call__(símismo, árbol):
        símismo.cnt = 0
        símismo.visit(árbol)
        return símismo.cnt



import random

import numpy as np

from tinamit.mod import VariablesMod, Variable, Modelo


class ModeloPrueba(Modelo):

    def __init__(símismo, unid_tiempo='años', nombre='prueba', dims=(1,)):
        símismo.unid_tiempo = unid_tiempo
        símismo.dims = dims
        super().__init__(nombre=nombre, variables=símismo._gen_vars())

    def incrementar(símismo, rebanada):
        símismo.variables.cambiar_vals({
            'Escala': símismo.variables['Escala'].obt_val() + rebanada.n_pasos,
        })

        super().incrementar(rebanada)

    def _gen_vars(símismo):
        m_inic = np.zeros(símismo.dims)
        return VariablesMod([
            Variable('Escala', unid=None, ingr=False, egr=True, inic=m_inic, líms=(0, None)),
            Variable('Vacío', unid=None, ingr=True, egr=False, inic=m_inic, líms=(0, None)),
            Variable('Vacío2', unid=None, ingr=True, egr=False, líms=(0, None), inic=[0, 0]),
        ])

    def unidad_tiempo(símismo):
        return símismo.unid_tiempo

    def paralelizable(símismo):
        return True


class ModeloAlea(Modelo):

    def __init__(símismo, unid_tiempo='años', nombre='alea'):
        símismo.unid_tiempo = unid_tiempo
        super().__init__(nombre=nombre, variables=símismo._gen_vars())

    def incrementar(símismo, rebanada):
        símismo.variables.cambiar_vals({
            'Escala': símismo.variables['Escala'].obt_val() + rebanada.n_pasos,
            'Aleatorio': random.random()
        })

        super().incrementar(rebanada)

    @staticmethod
    def _gen_vars():
        return VariablesMod([
            Variable('Escala', unid=None, ingr=False, egr=True, líms=(0, None)),
            Variable('Vacío', unid=None, ingr=True, egr=False, líms=(0, None)),
            Variable('Vacío2', unid=None, ingr=True, egr=False, líms=(0, None), inic=[0, 0]),
            Variable('Aleatorio', unid=None, ingr=False, egr=True, líms=(0, 1))
        ])

    def unidad_tiempo(símismo):
        return símismo.unid_tiempo

    def paralelizable(símismo):
        return True



from tinamit.config import _
from ..sintx.ec import Ecuación


class CalibradorEc(object):
    """
    Clase pariente para implementaciones de calibradores de ecuaciones.
    """

    def __init__(símismo, ec, paráms, nombre=None, dialecto='tinamït'):
        """

        Parameters
        ----------
        ec: str or Ec
            La ecuación para calibrar.
        paráms: list
            La lista de nombres de parámetros en la ecuación (variables que hay que calibrar).
        nombre: str
            El nombre del variable dependiente en la ecuación. Obligatorio si la ecuación no especifica variable
            independiente sí misma (p. ej., ``x * b + a`` en vez de ``y = x * b + a``.
        dialecto: str
            El dialecto de la ecuación. Puede ser ``tinamït`` o ``vensim``.
        """

        símismo.ec = ec if isinstance(ec, Ecuación) else Ecuación(ec, nombre=nombre, dialecto=dialecto)
        símismo.paráms = paráms

        # Asegurarse que se especificó el variable y.
        if símismo.ec.nombre is None:
            raise ValueError(_(
                'Debes especificar el nombre de la ecuación del variable dependiente, o con el parámetro'
                '`nombre`, o directamente en la ecuación, por ejemplo: "y = a*x ..."'
            ))

    def _gen_líms_paráms(símismo, líms_paráms):
        # Aplicar y formatear límites para parámetros sin límites
        líms_paráms = líms_paráms or {}
        for p in símismo.paráms:
            if p not in líms_paráms or líms_paráms[p] is None:
                líms_paráms[p] = (None, None)

        return líms_paráms

    def _extraer_vars(símismo):
        vars_ec = símismo.ec.variables()
        vars_x = [v for v in vars_ec if v not in símismo.paráms]
        var_y = símismo.ec.nombre

        return vars_x, var_y

    @staticmethod
    def _obt_datos(bd, vars_interés, corresp_vars):
        corresp_vars = corresp_vars or {}
        vars_bd = [v if v not in corresp_vars else corresp_vars[v] for v in vars_interés]
        datos = bd.interpolar(vars_bd)

        return datos.rename(
            {v: ll for ll, v in corresp_vars.items()}
        )

    def calibrar(símismo, bd, lugar=None, líms_paráms=None, ops=None, corresp_vars=None, ord_niveles=None):
        """
        Efectua la calibración.

        Parameters
        ----------
        bd: BD
            La base de datos con observaciones para los variables en la ecuación.
        lugar: Lugar
            El lugar cuyos sublugares hay que calibrar; si es ``None`` se calibrará la ecuación con todos
            los datos en ``bd`` sin tener su lugar en cuenta.
        líms_paráms: list
            Límites teoréticos para los parámetros.
        ops: dict
            Opciones que se pasarán directamente a la función de calibración.
        corresp_vars: dict
            Diccionario de correspondencia entre los nombres de los variables en ``bd`` y sus nombres en la ecuación.
        ord_niveles: list
            Desambiguación del orden de niveles.

        Returns
        -------
        dict
            Diccionario con las calibraciones de cada lugar.
        """

        raise NotImplementedError



from tinamit.config import _

from ._funcs import conv_fun


class VstrAPyMC3(object):
    def __init__(símismo, d_vars_pm, obs_x, í_datos, dialecto='tinamït'):
        símismo.d_vars_pm = d_vars_pm
        símismo.í_datos = í_datos
        símismo.obs_x = obs_x

        símismo.dialecto = dialecto

    def func(símismo, x):
        fun = conv_fun(x.children[0], símismo.dialecto, 'pm')  # Traducir la función a PyMC3
        args = [símismo.convertir(a) for a in
                x.children[1].children]  # Recursar a través de los argumentos de la función

        # Devolver la función dinámica
        return fun(*args)

    def neg(símismo, x):
        return -símismo.convertir(x.children[0])

    def var(símismo, x):
        v = str(x.children[0])
        try:
            if símismo.í_datos is None:
                return símismo.d_vars_pm[v]
            else:
                return símismo.d_vars_pm[v][símismo.í_datos]

        except KeyError:
            # Si el variable no es un parámetro calibrable, debe ser un valor observado
            try:
                return símismo.obs_x[v].values

            except KeyError:
                raise ValueError(_('El variable "{}" no es un parámetro, y no se encuentra'
                                   'en la base de datos observados tampoco.').format(v))

    def num(símismo, x):
        return float(x.children[0])

    def prod(símismo, x):
        # Recursar a través de los componentes del operador ahora. Es importante hacerlo antes de
        # las expresiones ``lambda``; sino se llamará `transformar()` sí mismo cada vez que se llama
        # la función generada.
        comp_1 = símismo.convertir(x.children[0])
        comp_2 = símismo.convertir(x.children[2])
        op = símismo.convertir(x.children[1])

        if op == '*':
            return comp_1 * comp_2
        elif op == '/':
            return comp_1 / comp_2
        else:
            raise ValueError(op)

    def op_prod(símismo, x):
        return str(x.children[0])

    def suma(símismo, x):
        comp_1 = símismo.convertir(x.children[0])
        comp_2 = símismo.convertir(x.children[2])
        op = símismo.convertir(x.children[1])

        if op == '+':
            return comp_1 + comp_2
        elif op == '-':
            return comp_1 - comp_2
        else:
            raise ValueError(op)

    def op_suma(símismo, x):
        return str(x.children[0])

    def pod(símismo, x):
        comp_1 = símismo.convertir(x.children[0])
        comp_2 = símismo.convertir(x.children[1])
        return comp_1 ** comp_2

    def comp(símismo, x):
        import pymc3 as pm

        comp_1 = símismo.convertir(x.children[0])
        comp_2 = símismo.convertir(x.children[2])
        op = símismo.convertir(x.children[1])

        if op == '>':
            return pm.math.gt(comp_1, comp_2)
        elif op == '<':
            return pm.math.lt(comp_1, comp_2)
        elif op == '>=':
            return pm.math.ge(comp_1, comp_2)
        elif op == '<=':
            return pm.math.le(comp_1, comp_2)
        elif op == '==':
            return pm.math.eq(comp_1, comp_2)
        elif op == '!=':
            return pm.math.neq(comp_1, comp_2)
        else:
            # Si no sabemos lo que tenemos, hay error.
            raise ValueError(op)

    def op_comp(símismo, x):
        return str(x.children[0])

    def convertir(símismo, árbol):
        if hasattr(símismo, árbol.data):
            return getattr(símismo, árbol.data)(árbol)
        else:
            raise TypeError(árbol.data)


class VstrAPy(object):
    def __init__(símismo, l_prms, dialecto):
        símismo.l_prms = l_prms
        símismo.dialecto = dialecto

    def func(símismo, x):
        fun = conv_fun(x.children[0], símismo.dialecto, 'python')  # Traducir la función a Python
        comp = [símismo.convertir(a) for a in
                x.children[1].children]  # Recursar a través de los argumentos de la función

        # Devolver la función dinámica
        return lambda p, vr: fun(*[a(p=p, vr=vr) for a in comp])

    def neg(símismo, x):
        comp = símismo.convertir(x.children[0])
        return lambda p, vr: -comp(p=p, vr=vr)

    def var(símismo, x):
        v = str(x.children[0])
        try:
            # Si el variable existe en la lista de parámetros, obtenerlo de `p`.
            í_var = símismo.l_prms.index(v)

            return lambda p, vr: p[í_var]

        except ValueError:
            # Si el variable no es un parámetro calibrable, debe ser un valor observado.
            return lambda p, vr: vr[v]

    def num(símismo, x):
        return lambda p, vr: float(x.children[0])

    def prod(símismo, x):
        # Recursar a través de los componentes del operador ahora. Es importante hacerlo antes de
        # las expresiones ``lambda``; sino se llamará `transformar()` sí mismo cada vez que se llama
        # la función generada.
        comp_1 = símismo.convertir(x.children[0])
        comp_2 = símismo.convertir(x.children[2])
        op = símismo.convertir(x.children[1])

        if op == '*':
            return lambda p, vr: comp_1(p=p, vr=vr) * comp_2(p=p, vr=vr)
        elif op == '/':
            return lambda p, vr: comp_1(p=p, vr=vr) / comp_2(p=p, vr=vr)
        else:
            raise ValueError(op)

    def op_prod(símismo, x):
        return str(x.children[0])

    def suma(símismo, x):
        comp_1 = símismo.convertir(x.children[0])
        comp_2 = símismo.convertir(x.children[2])
        op = símismo.convertir(x.children[1])

        if op == '+':
            return lambda p, vr: comp_1(p=p, vr=vr) + comp_2(p=p, vr=vr)
        elif op == '-':
            return lambda p, vr: comp_1(p=p, vr=vr) - comp_2(p=p, vr=vr)
        else:
            raise ValueError(op)

    def op_suma(símismo, x):
        return str(x.children[0])

    def pod(símismo, x):
        comp_1 = símismo.convertir(x.children[0])
        comp_2 = símismo.convertir(x.children[1])
        return lambda p, vr: comp_1(p=p, vr=vr) ** comp_2(p=p, vr=vr)

    def comp(símismo, x):
        comp_1 = símismo.convertir(x.children[0])
        comp_2 = símismo.convertir(x.children[2])
        op = símismo.convertir(x.children[1])

        if op == '>':
            return lambda p, vr: comp_1(p=p, vr=vr) > comp_2(p=p, vr=vr)
        elif op == '<':
            return lambda p, vr: comp_1(p=p, vr=vr) < comp_2(p=p, vr=vr)
        elif op == '>=':
            return lambda p, vr: comp_1(p=p, vr=vr) >= comp_2(p=p, vr=vr)
        elif op == '<=':
            return lambda p, vr: comp_1(p=p, vr=vr) <= comp_2(p=p, vr=vr)
        elif op == '==':
            return lambda p, vr: comp_1(p=p, vr=vr) == comp_2(p=p, vr=vr)
        elif op == '!=':
            return lambda p, vr: comp_1(p=p, vr=vr) != comp_2(p=p, vr=vr)
        else:
            # Si no sabemos lo que tenemos, hay error.
            raise ValueError(op)

    def op_comp(símismo, x):
        return str(x.children[0])

    def convertir(símismo, árbol):
        if hasattr(símismo, árbol.data):
            return getattr(símismo, árbol.data)(árbol)
        else:
            raise TypeError(árbol.data)


class VstrATx(object):
    def __init__(símismo, dialecto):
        símismo.dialecto = dialecto

    def func(símismo, x):
        args = [símismo.convertir(a) for a in x.children[1].children]
        if len(args) != 1:
            return str(x.children[0]) + '(' + ', '.join(args) + ')'
        elif len(args) == 1:
            return str(x.children[0]) + '(' + args[0] + ')'

    def neg(símismo, x):
        return '-' + símismo.convertir(x.children[0])

    def var(símismo, x):
        return str(x.children[0])

    def num(símismo, x):
        return str(x.children[0])

    def prod(símismo, x):
        return ''.join(símismo.convertir(a) for a in x.children)

    def op_prod(símismo, x):
        return str(x.children[0])

    def suma(símismo, x):
        return ''.join(símismo.convertir(a) for a in x.children)

    def op_suma(símismo, x):
        return str(x.children[0])

    def pod(símismo, x):
        return símismo.convertir(x.children[0]) + '^' + símismo.convertir(x.children[1])

    def comp(símismo, x):
        return ''.join(símismo.convertir(a) for a in x.children)

    def op_comp(símismo, x):
        return str(x.children[0])

    def convertir(símismo, árbol):
        if hasattr(símismo, árbol.data):
            return getattr(símismo, árbol.data)(árbol)
        else:
            raise TypeError(árbol.data)



"""
Código modificado de Jérôme Boisvert-Chouinard bajo la licencia MIT.
Code modified from Jérôme Boisvert-Chouinard under the MIT license.

The MIT License (MIT)

Copyright (c) 2014 Jerome Boisvert-Chouinard

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# !/bin/python

import itertools
##########
import sys
from ast import literal_eval

import numpy as np


# _anlzd_sntct.py
#
# Este módulo analiza archivos en formato de texto a base de una plantilla de la sintaxis y convierte entre varios
# formatos.
#
# Ejemplo de uso en línea de comanda:
#
# python _anlzd_sntct.py -r ejemplo_in.txt ejemplo1.tmpl
#
# Analizará ejemplo_in.txt y devolverá un diccionario de los parámetros analizados.
# Es únicamente útil así para la depuración.
#
# python _anlzd_sntct.py -c ejemplo_in.txt ejemplo1.tmpl ejemplo_egr.txt ejemplo2.tmpl
#
# Convertirá ejemplo_in.txt (en formato ejemplo1) a formato ejemplo2, y lo escribirá en ejemplo_egr.txt
#
# Ver la documentación por detalles acerca de la escritura de plantillas.
#
# jerome.boisvert-chouinard@mail.mcgill.ca
# Traducido y modificado por Julien Malard (julien.malard@mail.mcgill.ca)
#


def anlz_línea_l(línea, nombres_parámetros, dic_parámetros):
    if nombres_parámetros[0] == '#':
        dic_parámetros['#'].append(línea.strip())
    else:
        dic_parámetros[nombres_parámetros[0]] = línea.strip()
    return dic_parámetros


def construir_línea_l(nombres_parámetros, dic_parámetros):
    if nombres_parámetros[0] == '#':
        try:
            línea = dic_parámetros['#'].pop(0)
        except KeyError:
            línea = ''
        except IndexError:
            línea = ''
    else:
        línea = dic_parámetros[nombres_parámetros[0]]
    return línea + '\n'


def anlz_línea_f(línea, nombres_parámetros, anchura_cols, dic_parámetros):  # pragma: sin cobertura
    cursor = 0
    matriz_paráms = 0

    for i in range(len(anchura_cols)):
        if not matriz_paráms:
            if nombres_parámetros[i][0] != '*':
                dic_parámetros[nombres_parámetros[i]] = línea[cursor:cursor + anchura_cols[i]].strip('\n')
                cursor += anchura_cols[i]
            else:
                matriz_paráms = nombres_parámetros[i][1:]
                dic_parámetros[matriz_paráms] = []
        if matriz_paráms:
            dic_parámetros[matriz_paráms].append(línea[cursor:cursor + anchura_cols[i]].strip())
            cursor += anchura_cols[i]
    return dic_parámetros


def construir_línea_f(nombres_paráms, anchura_cols, dic_paráms):  # pragma: sin cobertura
    línea_egr = ''
    matriz_paráms = 0
    for i in range(len(anchura_cols)):
        if not matriz_paráms:
            nombre_parám = nombres_paráms[i]
            if nombre_parám[0] != '*':
                parám = dic_paráms[nombre_parám]
                parám = (anchura_cols[i] - len(parám)) * ' ' + parám
                línea_egr += parám
            else:
                matriz_paráms = nombre_parám.strip('*')
                lista_parám = [parameter for parameter in dic_paráms[matriz_paráms]]
        if matriz_paráms:
            parám = lista_parám.pop(0)
            parám = (anchura_cols[i] - len(parám)) * ' ' + parám
            línea_egr += parám
    return línea_egr + '  \n'


def anal_línea_d(línea, nombres_paráms, delim, dic_paráms):
    línea = línea.strip()
    if delim == 'W':
        valores = línea.split()
    else:
        valores = [value.strip() for value in línea.split(delim)]
    matriz_paráms = 0
    for i in range(len(valores)):
        if not matriz_paráms:
            try:
                if nombres_paráms[i][0] != '*':
                    dic_paráms[nombres_paráms[i]] = valores[i]
                else:
                    matriz_paráms = nombres_paráms[i][1:]
                    dic_paráms[matriz_paráms] = []
            except IndexError:
                return dic_paráms
        if matriz_paráms:
            dic_paráms[matriz_paráms].append(valores[i])
    return dic_paráms


def build_d_línea(nombres_paráms, delim, dic_paráms):
    valores = []
    # Si el separador es espacio blanco ('W'), emplear un espacio cono separador
    delim = '  ' if delim == 'W' else delim
    for nombreParám in nombres_paráms:
        if nombreParám[0] != '*':
            valores.append(str(dic_paráms[nombreParám]))
        else:
            valores += [str(x) for x in dic_paráms[nombreParám.strip('*')]]

    return delim.join(valores) + '  \n'


def anal_línea(línea, nombres_paráms, espec_línea, dic_paráms):
    if espec_línea[0] == 'L':
        anlz_línea_l(línea, nombres_paráms, dic_paráms)
    elif espec_línea[0] == 'F':
        anlz_línea_f(línea, nombres_paráms, espec_línea[1], dic_paráms)
    elif espec_línea[0] == 'D':
        anal_línea_d(línea, nombres_paráms, espec_línea[1], dic_paráms)
    return dic_paráms


def construir_línea(nombres_paráms, espec_línea, dic_paráms, dic_config, paráms_ent):
    if paráms_ent is not None:
        for p in dic_paráms:
            if p in paráms_ent:
                dic_paráms[p] = dic_paráms[p].astype(int)
    if espec_línea[0] == 'L':
        línea = construir_línea_l(nombres_paráms, dic_paráms)
    elif espec_línea[0] == 'F':
        línea = construir_línea_f(nombres_paráms, espec_línea[1], dic_paráms)
    elif espec_línea[0] == 'D':
        línea = build_d_línea(nombres_paráms, espec_línea[1], dic_paráms)
    else:
        raise ValueError(espec_línea[0])
    if 'CSVPAD' in dic_config.keys():
        diff = dic_config['CSVPAD'] - 1 - línea.count(',')
        if diff > 0:
            línea = línea.strip() + ',' * diff + '\n'
    return línea


def _crear_matriz_de_zeros(*dims):
    if len(dims) == 1:
        return [0] * dims[0]
    else:
        matr = []
        for i in range(dims[0]):
            matr.append(_crear_matriz_de_zeros(*dims[1:]))
        return matr


def leer_archivo(nombre_arch_contenido, nombre_arch_plantilla, paráms_ent=None):
    if paráms_ent is None:
        paráms_ent = []

    dic_paráms = {'#': []}
    config_dic = {}
    with open(nombre_arch_contenido, 'r') as archContenido, open(nombre_arch_plantilla, 'r') as archPlantilla:
        for línea_plantilla in archPlantilla:

            if línea_plantilla[0] == '!':
                config_tupla = literal_eval(línea_plantilla[1:].strip())
                config_dic[config_tupla[0]] = config_tupla[1]
            elif línea_plantilla[0] != '#':
                template_tuple = literal_eval(línea_plantilla.strip())
                if template_tuple[0][0:4] != 'FOR[':
                    línea_contenido = archContenido.readline().strip('\n')
                    anal_línea(línea_contenido, template_tuple[1], template_tuple[0], dic_paráms)
                else:
                    dims = [i.strip(']') for i in template_tuple[0].split('[')][1:]
                    for i in range(len(dims)):
                        try:
                            dims[i] = int(dims[i])
                        except ValueError:
                            dims[i] = int(dic_paráms[dims[i]])
                    tuple_espec_línea = template_tuple[1]
                    for especLínea in tuple_espec_línea:

                        for nombre_parám in especLínea[1]:
                            nombre_parám = nombre_parám.strip('*')
                            dic_paráms[nombre_parám] = _crear_matriz_de_zeros(*dims)

                    iterdims = [range(dim) for dim in dims]
                    for indices in itertools.product(*iterdims):
                        dic_temp = {}
                        for especLínea in tuple_espec_línea:
                            línea_contenido = archContenido.readline().strip('\n')
                            anal_línea(línea_contenido, especLínea[1], especLínea[0], dic_temp)
                            for nombre_parám in especLínea[1]:
                                nombre_parám = nombre_parám.strip("*")
                                d = dic_paráms[nombre_parám]
                                for i in indices[:-1]:
                                    d = d[i]
                                d[indices[-1]] = dic_temp[nombre_parám]

    for k, v in dic_paráms.items():
        if isinstance(v, list):
            if k in paráms_ent:
                dic_paráms[k] = np.array(v).astype(int)
            else:
                dic_paráms[k] = np.array(v).astype(float)

    return dic_paráms


def escribir_archivo(dic_paráms, nombre_arch_contenido, nombre_arch_plantilla, paráms_ent=None):
    # k=1
    dic_config = {}
    with open(nombre_arch_contenido, 'w') as archContenido, open(nombre_arch_plantilla, 'r') as arch_plantilla:
        for templatelínea in arch_plantilla:
            if templatelínea[0] == '!':
                tupla_config = literal_eval(templatelínea[1:].strip())
                dic_config[tupla_config[0]] = tupla_config[1]
            elif templatelínea[0] != '#':
                tupla_plantilla = literal_eval(templatelínea.strip())
                if tupla_plantilla[0][0:4] != 'FOR[':
                    archContenido.write(
                        construir_línea(tupla_plantilla[1], tupla_plantilla[0], dic_paráms, dic_config, paráms_ent)
                    )
                else:
                    dims = [i.strip(']') for i in tupla_plantilla[0].split('[')][1:]
                    for i in range(len(dims)):
                        try:
                            dims[i] = int(dims[i])
                        except ValueError:
                            dims[i] = int(dic_paráms[dims[i]])
                    línea_espec_tuple = tupla_plantilla[1]
                    iterdims = [range(dim) for dim in dims]
                    for indices in itertools.product(*iterdims):
                        dic_temp = {}
                        for especLínea in línea_espec_tuple:
                            for nombre_parám in especLínea[1]:
                                nombre_parám = nombre_parám.strip('*')
                                p_d = dic_paráms[nombre_parám]
                                for i in indices:
                                    p_d = p_d[i]

                                dic_temp[nombre_parám] = p_d

                            archContenido.write(
                                construir_línea(especLínea[1], especLínea[0], dic_temp, dic_config, paráms_ent))
    return nombre_arch_contenido


def central(*args):  # pragma: sin cobertura
    if args[1] == '-r':
        dic_paráms = leer_archivo(args[2], args[3])
    elif args[1] == '-c':
        dic_paráms = leer_archivo(args[2], args[3])
        escribir_archivo(dic_paráms, args[4], args[5])
    return 0


# El variable __name___ es igual a __main__ cuando el programa se llama desde la línea de comanda.
# Así, se corre la función `central` cuando se llama desde la línea de comanda, pero no cuando se importa el módulo.
if __name__ == '__main__':
    central(*sys.argv)



import datetime as ft

import numpy as np
import pandas as pd
import xarray as xr
from tinamit.config import _
from tinamit.tiempo.tiempo import TiempoCalendario, Tiempo

from .var import Variable


class Extern(object):
    """
    Datos externos para simulaciones.
    """

    def __init__(símismo, vals_vars, interpol=True):
        """

        Parameters
        ----------
        vals_vars: dict[str, xr.DataArray]
            Diccionario de los valores de los variables.
        interpol: bool
            Si se puede interpolar los datos.
        """
        símismo.interpol = interpol
        símismo._vals = vals_vars

    def obt_vals(símismo, t, var=None):
        """
        Devuelve los valores de uno o más variables.

        Parameters
        ----------
        t
            El tiempo de interés.
        var: str or Variable or list
            Los variables de interés.

        Returns
        -------
        dict[str, xr.DataArray]
        """
        vals = símismo._vals

        if var is not None:
            if isinstance(var, (str, Variable)):
                var = [var]
            vals = {vr: vl for vr, vl in vals.items() if vr in var}

        vals = {vr: símismo._obt_a_t(vl, t, símismo.interpol) for vr, vl in vals.items()}

        return {vr: vl for vr, vl in vals.items() if not np.all(np.isnan(vl))}

    @staticmethod
    def _obt_a_t(m_xr, t, interpol):
        m_xr = m_xr.unstack()

        t_rel = relativizar_eje(m_xr, t)
        if interpol and m_xr.sizes[_('fecha')] > 1:
            return m_xr.interp(**{_('fecha'): t_rel}).dropna(_('fecha'))
        try:
            return m_xr.reindex({_('fecha'): t_rel}).dropna(_('fecha'))
        except (KeyError, IndexError):
            return np.nan


def gen_extern(datos, interpol=True):
    """
    Transforma datos en objeto :class:`~tinamit.extern.Extern`.

    Parameters
    ----------
    datos: Extern or pd.DataFrame or xr.Dataset or dict
        Los datos.
    interpol: bool
        Si se pueden interpolar los datos.

    Returns
    -------
    Extern

    """
    if isinstance(datos, Extern) or datos is None:
        return datos

    if isinstance(datos, pd.DataFrame):
        datos = datos.to_xarray().rename({'index': _('fecha')})

    if isinstance(datos, xr.Dataset):
        return Extern({vr: datos[vr] for vr in datos.data_vars}, interpol)
    elif isinstance(datos, dict):
        return Extern({vr: _a_matr_xr(vl) for vr, vl in datos.items()})
    raise TypeError(type(datos))


def _a_matr_xr(val):
    if isinstance(val, xr.DataArray):
        return val
    else:
        if not isinstance(val, np.ndarray):
            val = np.array([val])
        return xr.DataArray(
            val.reshape((1, *val.shape)),
            coords={_('fecha'): [0]},
            dims=[_('fecha'), *('x_' + str(i) for i in range(len(val.shape)))]
        )


def relativizar_eje(ref, otro):
    índ_ref = _obt_índ(ref)
    if isinstance(otro, TiempoCalendario):
        índ_otro = pd.to_datetime([otro.fecha()]) if isinstance(índ_ref, pd.DatetimeIndex) else pd.Index([otro.í])
    elif isinstance(otro, Tiempo):
        índ_otro = pd.Index([otro.í])
    else:
        índ_otro = _obt_índ(otro)

    if isinstance(índ_ref, pd.DatetimeIndex):
        if isinstance(índ_otro, pd.DatetimeIndex):
            return índ_otro
        raise TypeError

    if isinstance(índ_otro, pd.DatetimeIndex):
        return pd.to_datetime(índ_ref, unit='D', origin=índ_otro[0])

    return índ_otro


def _obt_índ(dts):
    if isinstance(dts, pd.Index):
        return dts
    if isinstance(dts, pd.DataFrame):
        return dts.index
    if isinstance(dts, (xr.Dataset, xr.DataArray)):
        return dts[_('fecha')].to_pandas().index
    if isinstance(dts, (float, int)):
        return pd.Index([dts])
    if isinstance(dts, np.ndarray):
        return pd.Index(dts)
    if isinstance(dts, (np.datetime64, str, ft.datetime, ft.date)):
        return pd.to_datetime([dts])

    raise TypeError(type(dts))



import unittest

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import xarray as xr

from pruebas.recursos.mod.prueba_mod import ModeloPrueba
from tinamit.config import trads
from tinamit.tiempo.tiempo import EspecTiempo
from tinamit.mod import OpsSimulGrupoCombin, OpsSimulGrupo

_ = trads.trad


class TestSimular(unittest.TestCase):
    def test_simular_t_int(símismo):
        mod = ModeloPrueba()
        res = mod.simular(t=10)['Escala']
        npt.assert_equal(res.vals, np.arange(11).reshape((11, 1)))

    def test_simular_t_fecha(símismo):
        mod = ModeloPrueba(unid_tiempo='días')
        res = mod.simular(t=EspecTiempo(10, '2001-01-01'))['Escala']
        pdt.assert_index_equal(
            res.vals.indexes[_('fecha')], pd.date_range('2001-01-01', periods=11), check_names=False
        )

    def test_paso(símismo):
        mod = ModeloPrueba()
        res = mod.simular(t=EspecTiempo(10, tmñ_paso=2))['Escala']
        npt.assert_equal(res.vals, np.arange(21, step=2).reshape((11, 1)))
        npt.assert_equal(res.vals.indexes[_('fecha')], np.arange(21, step=2))

    def test_paso_inválido(símismo):
        mod = ModeloPrueba()
        with símismo.assertRaises(ValueError):
            mod.simular(EspecTiempo(100, tmñ_paso=0))

    def test_guardar_cada(símismo):
        mod = ModeloPrueba()
        res = mod.simular(t=EspecTiempo(10, guardar_cada=2))['Escala']
        npt.assert_equal(res.vals, np.arange(11, step=2).reshape((6, 1)))
        npt.assert_equal(res.t.eje(), np.arange(11, step=2))

    def test_guardar_cada_con_fecha(símismo):
        mod = ModeloPrueba(unid_tiempo='días')
        res = mod.simular(t=EspecTiempo(10, '2001-01-01', guardar_cada=2))['Escala']
        npt.assert_equal(res.vals, np.arange(11, step=2).reshape((6, 1)))
        pdt.assert_index_equal(
            res.vals.indexes[_('fecha')], pd.date_range('2001-01-01', periods=6, freq='2D'), check_names=False
        )


class TestSimularGrupo(unittest.TestCase):

    def test_simular_grupo(símismo):
        vals_extern = [{'Escala': 1}, {'Escala': 2}]
        mod = ModeloPrueba()
        ops = OpsSimulGrupo(5, extern=vals_extern)
        res = mod.simular_grupo(ops)
        for corr, vl in zip(res.values(), vals_extern):
            símismo.assertEqual(corr['Escala'].vals.values[0], vl['Escala'])

    def test_simular_grupo_combin(símismo):
        mod = ModeloPrueba()
        ops = OpsSimulGrupoCombin([5, 6], extern=[{'Escala': 1}, {'Escala': 2}])
        res = mod.simular_grupo(ops)
        símismo.assertEqual(len(res), 4)

    def test_simular_grupo_tmñ_erróneo(símismo):
        with símismo.assertRaises(ValueError):
            OpsSimulGrupo([5, 6, 7], extern=[{'Escala': 1}, {'Escala': 2}])

    def test_simular_paralelo(símismo):
        mod = ModeloPrueba()
        ops = OpsSimulGrupoCombin(5, extern=[{'Escala': 1}, {'Escala': 2}])
        sin_paralelo = mod.simular_grupo(ops)
        con_paralelo = mod.simular_grupo(ops, paralelo=True)
        símismo.assertEqual(sin_paralelo, con_paralelo)

    def test_simular_grupo_con_lista_nombres(símismo):
        mod = ModeloPrueba()
        nombres = ['corrida 1', 'corrida 2']
        ops = OpsSimulGrupo(5, extern=[{'Escala': 1}, {'Escala': 2}], nombre=nombres)
        res = mod.simular_grupo(ops)
        símismo.assertSetEqual(set(nombres), set(res))


class TestSimulConDatos(unittest.TestCase):

    @staticmethod
    def _simul_con_extern(extern, ref, var_ref='Vacío', fecha=True):
        f_inic = '2000-01-01' if fecha else None
        res = ModeloPrueba(unid_tiempo='días').simular(t=EspecTiempo(10, f_inic=f_inic), extern=extern)
        npt.assert_equal(res[var_ref].vals.values.reshape(np.array(ref).shape), ref)

    def test_t_fecha_extern_núm(símismo):
        símismo._simul_con_extern({'Vacío': 4}, ref=np.full(11, 4))

    def test_t_fecha_extern_matr(símismo):
        símismo._simul_con_extern({'Vacío2': np.arange(2)}, np.tile(np.arange(2), (11, 1)), var_ref='Vacío2')

    def test_t_fecha_extern_pd_num(símismo):
        extern = pd.DataFrame(data={'Vacío': np.arange(5)}, index=np.arange(10, step=2))
        símismo._simul_con_extern(extern, ref=[0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4, 4])

    def test_t_fecha_extern_pd_fecha(símismo):
        extern = pd.DataFrame(data={'Vacío': np.arange(4)}, index=pd.date_range('2000-01-02', '2000-01-05'))
        símismo._simul_con_extern(extern, ref=[0, 0, 1, 2, 3, 3, 3, 3, 3, 3, 3])

    def test_t_fecha_extern_xr_fecha(símismo):
        extern = {
            'Vacío': xr.DataArray(
                np.arange(11), coords={_('fecha'): pd.date_range('2000-01-01', periods=11)}, dims=[_('fecha')]
            )
        }
        símismo._simul_con_extern(extern, ref=np.arange(11))

    def test_t_fecha_extern_xr_num(símismo):
        extern = {'Vacío': xr.DataArray(np.arange(11), coords={_('fecha'): np.arange(11)}, dims=[_('fecha')])}
        símismo._simul_con_extern(extern, ref=np.arange(11))

    def test_t_fecha_extern_bd_xr(símismo):
        extern = xr.Dataset({'Vacío': (_('fecha'), np.arange(11))}, coords={_('fecha'): np.arange(11)})
        símismo._simul_con_extern(extern, ref=np.arange(11))

    def test_t_numérico(símismo):
        extern = pd.DataFrame(data={'Vacío': np.arange(5)}, index=np.arange(10, step=2))
        símismo._simul_con_extern(extern, fecha=False, ref=[0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4, 4])

    def test_t_numérico_extern_fecha(símismo):
        extern = pd.DataFrame(data={'Vacío': np.arange(4)}, index=pd.date_range('2000-01-02', '2000-01-05'))
        with símismo.assertRaises(TypeError):
            ModeloPrueba(unid_tiempo='días').simular(t=EspecTiempo(10), extern=extern)



import os
import unittest

import numpy.testing as npt
import xarray.testing as xrt

from pruebas.recursos.mod.prueba_mod import ModeloAlea
from pruebas.test_mds import generar_modelos_prueba, limpiar_mds
from tinamit.conect import Conectado, SuperConectado
from tinamit.envolt.bf import gen_bf
from tinamit.mod import OpsSimulGrupo
from tinamit.unids import trads

dir_act = os.path.split(__file__)[0]
arch_bf = os.path.join(dir_act, 'recursos/bf/prueba_bf.py')
arch_mds = os.path.join(dir_act, 'recursos/mds/prueba_senc_mdl.mdl')


# Comprobar Conectado
class TestConectado(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Preparar los modelos genéricos necesarios para las pruebas.
        """

        # Generar las instancias de los modelos individuales y conectados
        cls.mods_mds = generar_modelos_prueba()

        cls.modelos = {ll: Conectado(bf=gen_bf(arch_bf), mds=mod) for ll, mod in cls.mods_mds.items()}

        # Agregar traducciones necesarias.
        trads.agregar_trad('year', 'año', leng_trad='es', leng_orig='en')
        trads.agregar_trad('month', 'mes', leng_trad='es', leng_orig='en')

        # Conectar los modelos
        for mod in cls.modelos.values():
            mod.conectar(var_mds='Lluvia', var_bf='Lluvia', mds_fuente=False)
            mod.conectar(var_mds='Lago', var_bf='Lago', mds_fuente=True)

    def test_reinic_vals(símismo):
        """
        Asegurarse que los variables se reinicializen correctamente después de una simulación.
        """

        for ll, mod in símismo.modelos.items():
            with símismo.subTest(mod=ll):
                egr1 = mod.simular(100, vars_interés=list(mod.variables))
                egr2 = mod.simular(100, vars_interés=list(mod.variables))

                for r1, r2 in zip(egr1, egr2):
                    xrt.assert_equal(r1.vals, r2.vals)

    def test_intercambio_de_variables(símismo):
        """
        Asegurarse que valores intercambiados tengan valores iguales en los resultados de ambos modelos.
        """

        for ll, mod in símismo.modelos.items():
            with símismo.subTest(mod=ll):
                res = mod.simular(100)

                npt.assert_array_equal(
                    res['bf']['Lluvia'].vals.values.flatten(), res['mds']['Lluvia'].vals.values.flatten()
                )
                npt.assert_array_equal(
                    res['bf']['Lago'].vals.values.flatten(), res['mds']['Lago'].vals.values.flatten()
                )

    def test_simular_grupo(símismo):
        """
        Comprobar que simulaciones en grupo den el mismo resultado que las mismas simulaciones individuales.
        """

        for ll, mod in símismo.modelos.items():
            with símismo.subTest(mod=ll):
                t_final = 10
                ref = {
                    'lago_%i' % i: mod.simular(t_final, extern={'Nivel lago inicial': i}, vars_interés='Lago')
                    for i in [50, 2000]
                }

                ops = OpsSimulGrupo(
                    t_final, extern=[{'Nivel lago inicial': 50}, {'Nivel lago inicial': 2000}],
                    nombre=['lago_50', 'lago_2000'],
                    vars_interés='Lago'
                )
                resultados = mod.simular_grupo(ops)

                for c in ref:
                    npt.assert_allclose(
                        ref[c]['bf']['Lago'].vals.values.flatten(),
                        resultados[c]['bf']['Lago'].vals.values.flatten(), rtol=0.001
                    )

    @classmethod
    def tearDownClass(cls):
        """
        Limpiar todos los archivos temporarios.
        """

        limpiar_mds()


class Test3ModelosConectados(unittest.TestCase):
    """
    Comprobar 3+ modelos conectados.
    """

    def test_conexión_horizontal(símismo):
        """
        Comprobar que la conexión de variables se haga correctamente con 3 submodelos conectados directamente
        en el mismo :class:`SuperConectado`.
        """

        # Crear los 3 modelos
        mod_1 = ModeloAlea(nombre='modelo 1')
        mod_2 = ModeloAlea(nombre='modelo 2')
        mod_3 = ModeloAlea(nombre='modelo 3')

        # El Conectado
        conectado = SuperConectado([mod_1, mod_2, mod_3])

        # Conectar variables entre dos de los modelos por el intermediario del tercero.
        conectado.conectar_vars(
            var_fuente='Aleatorio', modelo_fuente='modelo 1', var_recip='Vacío', modelo_recip='modelo 2'
        )
        conectado.conectar_vars(
            var_fuente='Vacío', modelo_fuente='modelo 2', var_recip='Vacío', modelo_recip='modelo 3'
        )

        # Simular
        res = conectado.simular(10, vars_interés=[mod_1.variables['Aleatorio'], mod_3.variables['Vacío']])

        # Comprobar que los resultados son iguales.
        xrt.assert_equal(res['modelo 1']['Aleatorio'].vals, res['modelo 3']['Vacío'].vals)

    def test_conexión_jerárquica(símismo):
        """
        Comprobar que 3 modelos conectados de manera jerárquica a través de 2 :class:`SuperConectados` funcione
        bien. No es la manera más fácil o directa de conectar 3+ modelos, pero es importante que pueda funcionar
        esta manera también.
        """

        # Los tres modelos
        mod_1 = ModeloAlea(nombre='modelo 1')
        mod_2 = ModeloAlea(nombre='modelo 2')
        mod_3 = ModeloAlea(nombre='modelo 3')

        # El primer Conectado
        conectado_sub = SuperConectado(nombre='sub', modelos=[mod_1, mod_2])
        conectado_sub.conectar_vars(
            var_fuente='Aleatorio', modelo_fuente='modelo 1', var_recip='Vacío', modelo_recip='modelo 2'
        )

        # El segundo Conectado
        conectado = SuperConectado([conectado_sub, mod_3])
        conectado.conectar_vars(
            var_fuente=mod_2.variables['Vacío'], var_recip='Vacío', modelo_recip='modelo 3'
        )

        # Correr la simulación
        res = conectado.simular(10, vars_interés=[mod_1.variables['Aleatorio'], mod_3.variables['Vacío']])

        # Verificar que los resultados sean iguales.
        xrt.assert_equal(res['sub']['modelo 1']['Aleatorio'].vals, res['modelo 3']['Vacío'].vals)



import datetime as ft
import os
import unittest

import numpy as np
import numpy.testing as npt
import pandas as pd
import tinamit.datos.fuente as fnt
import xarray as xr
import xarray.testing as xrt
from tinamit.config import _
from tinamit.datos.bd import BD

dir_act = os.path.split(__file__)[0]
arch_datos1 = os.path.join(dir_act, 'recursos/datos/datos.csv')
arch_datos2 = os.path.join(dir_act, 'recursos/datos/datos2.csv')


class TestGeneraFuentes(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        dic_datos = {
            'lugar': [708, 708, 7],
            'fecha': ['2000', '2002', '2000'],
            'var_completo': [1, 2, 2],
            'var_incompleto': [np.nan, 10, 20],
            'temp': [1, 1, 1]
        }

        cls.bds = {
            'pandas': fnt.FuentePandas(pd.DataFrame(dic_datos), 'Pandas'),
            'csv': fnt.FuenteCSV(arch_datos1, 'CSV'),
            'dic': fnt.FuenteDic(dic_datos, 'Dic'),
            'xarray': fnt.FuenteBaseXarray(xr.Dataset(
                {
                    'var_completo': ('n', dic_datos['var_completo']),
                    'var_incompleto': ('n', dic_datos['var_incompleto']),
                    'fecha': ('n', dic_datos['fecha']),
                    'lugar': ('n', dic_datos['lugar'])
                }
            ), 'Xarray', fechas='fecha', lugares='lugar'
            )
        }

    def test_obt_datos(símismo):
        for nmb, bd in símismo.bds.items():
            with símismo.subTest(bd=nmb):
                datos = bd.obt_vals(['var_completo'])
                npt.assert_array_equal(datos['var_completo'], [1, 2, 2])

    def test_obt_datos_faltan(símismo):
        for nmb, bd in símismo.bds.items():
            with símismo.subTest(bd=nmb):
                datos = bd.obt_vals(['var_incompleto'])
                npt.assert_array_equal(datos['var_incompleto'], [np.nan, 10, 20])


class TestFuentes(unittest.TestCase):

    def test_fechas_columna(símismo):
        fechas = ['1947-8-14', '1947-8-15']
        fuente = fnt.FuentePandas(pd.DataFrame({'a': [1, 2], 'f': fechas}), 'prueba', fechas='f')
        lista_fechas_igual(fuente.obt_vals('a')[_('fecha')].values, fechas)

    def test_fechas_texto_año(símismo):
        fuente = fnt.FuentePandas(pd.DataFrame({'a': [1, 2]}), 'prueba', fechas='1966')
        lista_fechas_igual(fuente.obt_vals('a')[_('fecha')].values, [ft.date(1966, 1, 1)] * 2)

    def test_fechas_ft(símismo):
        fecha = ft.date(1966, 1, 1)
        fuente = fnt.FuentePandas(pd.DataFrame({'a': [1, 2]}), 'prueba', fechas=fecha)
        lista_fechas_igual(fuente.obt_vals('a')[_('fecha')].values, [fecha] * 2)

    def test_fecha_texto(símismo):
        fecha = '1966-7-21'
        fuente = fnt.FuentePandas(pd.DataFrame({'a': [1, 2]}), 'prueba', fechas=fecha)
        lista_fechas_igual(fuente.obt_vals('a')[_('fecha')].values, [fecha] * 2)

    def test_fechas_error(símismo):
        with símismo.assertRaises(ValueError):
            fnt.FuentePandas(pd.DataFrame({'a': [1, 2]}), 'prueba', fechas='No soy una columna').obt_vals('a')

    def test_lugares_columna(símismo):
        lugar = ['1', '2']
        fuente = fnt.FuentePandas(pd.DataFrame({'a': [1, 2], 'l': lugar}), 'prueba', lugares='l')
        lugares_fuente = fuente.obt_vals('a').lugar.values.tolist()
        símismo.assertListEqual(lugares_fuente, lugar)

    def test_lugares_texto(símismo):
        fuente = fnt.FuentePandas(pd.DataFrame({'a': [1, 2]}), 'prueba', lugares='708')
        lugares_fuente = fuente.obt_vals('a')[_('lugar')].values.tolist()
        símismo.assertEqual(lugares_fuente, ['708', '708'])


class TestBD(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        fnt1 = fnt.FuenteCSV(arch_datos1, nombre='prueba 1', fechas='fecha', lugares='lugar')
        fnt2 = fnt.FuenteCSV(arch_datos2, nombre='prueba 1', fechas='fecha', lugares='lugar')
        cls.bd = BD(fuentes=[fnt1, fnt2])

    def test_obt_datos_1_var(símismo):
        res = símismo.bd.obt_vals('var_completo')
        símismo.assertSetEqual(_a_conj(res, 'lugar'), {'701', '708', '7'})
        símismo.assertEqual(res.name, 'var_completo')

    def test_obt_datos_multi_vars(símismo):
        res = símismo.bd.obt_vals(['var_completo', 'var_incompleto'])
        símismo.assertSetEqual(_a_conj(res, 'lugar'), {'701', '708', '7'})
        símismo.assertIn('var_completo', res)
        símismo.assertIn('var_incompleto', res)

    def test_obt_datos_de_lugar(símismo):
        res = símismo.bd.obt_vals('var_completo', lugares='708')
        símismo.assertTrue(_a_conj(res, 'lugar') == {'708'})

    def test_obt_datos_fecha_única(símismo):
        res = símismo.bd.obt_vals('var_completo', fechas='2000')
        símismo.assertTrue(fechas_en_rango(res[_('fecha')], ['2000-1-1', '2000-12-31']))
        npt.assert_equal(res.values, [1, 2])

    def test_obt_datos_fecha_lista(símismo):
        res = símismo.bd.obt_vals('var_completo', fechas=['2000', '2002'])
        símismo.assertTrue(fechas_en_lista(res[_('fecha')], ['2000-01-01', '2002-01-01']))

    def test_obt_datos_fecha_rango(símismo):
        res = símismo.bd.obt_vals('var_completo', fechas=('2000', '2002-6-1'))
        símismo.assertTrue(fechas_en_rango(res[_('fecha')], ('2000-1-1', '2002-6-1')))

    def test_interpol_no_necesaria(símismo):
        res = símismo.bd.interpolar('var_completo', fechas=('2000', '2002-1-1'), lugares='708')

        ref = símismo.bd.obt_vals('var_completo', lugares='708')
        xrt.assert_equal(
            res.where(res[_('fecha')].isin(pd.to_datetime(('2000', '2002-1-1'))), drop=True),
            ref.where(ref[_('fecha')].isin(pd.to_datetime(('2000', '2002-1-1'))), drop=True)
        )

    def test_interpol_con_fecha_única(símismo):
        res = símismo.bd.interpolar(['var_incompleto', 'var_completo'], fechas='2001')

        # Verificar interpolaciones
        vals = res.where(
            np.logical_and(res[_('lugar')] == '7', res[_('fecha')] == np.datetime64('2001-01-01')),
            drop=True
        )
        npt.assert_allclose(vals['var_completo'].values, 2.500684)
        npt.assert_allclose(vals['var_incompleto'].values, 23.3394, rtol=0.001)

    def test_interpol_con_fecha_rango(símismo):
        l_vars = ['var_incompleto', 'var_completo']
        res = símismo.bd.interpolar(l_vars, fechas=('2000', '2002-6-1'))
        npt.assert_allclose(res['var_incompleto'].sel(lugar='7'), [20, 28.05], rtol=0.01)


def _a_conj(bd, var):
    return set(np.unique(bd[var].values).tolist())


def lista_fechas_igual(l1, l2):
    lista_ls = [l1, l2]
    for i, l in enumerate(lista_ls):
        if not isinstance(l, np.ndarray):
            lista_ls[i] = np.array(pd.to_datetime(l), dtype='datetime64')
        else:
            if not np.issubdtype(l.dtype, np.datetime64):
                lista_ls[i] = l.astype('datetime64')

    npt.assert_array_equal(lista_ls[0], lista_ls[1])


def fechas_en_lista(f, l):
    l = np.array(pd.to_datetime(l), dtype='datetime64')

    return np.all(f.isin(l))


def fechas_en_rango(f, r):
    r = np.array(pd.to_datetime(r), dtype='datetime64')

    return np.all(f <= r[1]) and np.all(f >= r[0])



import os
from subprocess import run

from tinamit import __file__ as dir_tinamit

leng_orig = 'es'
lenguas = {'ms', 'fr', 'ta', 'ur', 'nl'}

# El directorio de tinamit
dir_local = os.path.split(os.path.realpath(__file__))[0]

dir_base, pq = os.path.split(os.path.realpath(os.path.split(dir_tinamit)[0]))

dir_fuente = os.path.join(dir_base, pq, '_local', '_fuente')
lengs_ya = [x for x in os.listdir(dir_fuente) if os.path.isdir(os.path.join(dir_fuente, x))]
lengs_nuevas = lenguas.difference(lengs_ya)
lenguas.update(lengs_ya)

run('python setup.py extract_messages', cwd=dir_base)

for l in lengs_nuevas:
    run('python setup.py init_catalog -l {}'.format(l), cwd=dir_base)

for l in lengs_ya:
    run('python setup.py update_catalog -l {}'.format(l), cwd=dir_base)

# Mandar cambios locales al servidor Zanata
print('Mandando traducciones actualizadas localmente a Zanata...')
run('zanata po push --copytrans --import-po', input=b'y', cwd=dir_local)

# Traemos traducciones de Transifex y las mandamos a Zanata.
print('Actualizando con Transifex...')
run('tx pull -a -f', cwd=dir_fuente)
print('Mandando traducciones de Transifex a Zanata...')
run('zanata po push --copytrans --import-po', input=b'y', cwd=dir_local)

# Traemos las traducciones más recientes de Zanata
print('Verificando las traducciones más recientes en Zanata...')
run('zanata po pull', cwd=dir_local)

# Mandar los documentos de traducciones actualizados al servidor Transifex
print('Mandando todo a Transifex también...')
run('tx push -s -t', cwd=dir_local)

# Ver las estadísticas
print('Pidiendo estadísticas recientes de traducción (de Zanata)...')
run('zanata stats', cwd=dir_local)

# Compilar las traducciones actualizadas
run('python setup.py compile_catalog', cwd=dir_base)



import csv
import datetime as ft
import os

import numpy as np
import pandas as pd
import xarray as xr
from tinamit.config import _
from tinamit.cositas import detectar_codif
from எண்ணிக்கை import எண்ணுக்கு as எ


class Fuente(object):
    """
    La clase pariente para fuentes de datos.
    """

    def __init__(símismo, nombre, variables, lugares=None, fechas=None):
        símismo.nombre = nombre

        símismo.variables = [vr for vr in variables if vr not in [lugares, fechas]]

        símismo._equiv_nombres = {}

        if fechas is None:
            fechas = pd.NaT
        símismo.n_obs = símismo._vec_var(símismo.variables[0], tx=True).size

        símismo.lugares = símismo._obt_lugar(lugares)
        símismo.fechas = símismo._obt_fecha(fechas)

    def obt_vals(símismo, vars_interés, lugares=None, fechas=None):
        vars_interés = vars_interés or símismo.variables

        coords = {_('lugar'): ('n', símismo.lugares), _('fecha'): ('n', símismo.fechas)}
        if isinstance(vars_interés, str):
            vals = xr.DataArray(
                símismo._vec_var(símismo._resolver_nombre(vars_interés)),
                coords=coords, dims='n', name=vars_interés
            )
        else:
            vals = xr.Dataset(
                {vr: ('n', símismo._vec_var(símismo._resolver_nombre(vr)))
                 for vr in vars_interés if vr in símismo.variables
                 },
                coords=coords
            )
        vals = vals.set_index(n=['lugar', 'fecha'])
        return símismo._filtrar_lugares(símismo._filtrar_fechas(vals, fechas), lugares)

    def _obt_lugar(símismo, lugares):
        lugares = lugares or ''
        if isinstance(lugares, str):
            try:
                lugares = símismo._vec_var(lugares, tx=True)
            except KeyError:
                return np.full(símismo.n_obs, lugares)

        return lugares

    def _obt_fecha(símismo, fechas):
        if isinstance(fechas, str):
            try:
                fechas = símismo._vec_var(fechas, tx=True)
            except KeyError:
                pass
            fechas = pd.to_datetime(fechas)
        if isinstance(fechas, pd.Timestamp):
            fechas = pd.to_datetime(np.full(símismo.n_obs, fechas.to_datetime64()))
        elif isinstance(fechas, (ft.date, ft.datetime)):
            fechas = pd.to_datetime(np.full(símismo.n_obs, fechas).astype(ft.datetime))

        return fechas

    def equiv_nombre(símismo, var, equiv):
        símismo._equiv_nombres[equiv] = var

    def _resolver_nombre(símismo, var):
        try:
            símismo._equiv_nombres[var]
        except KeyError:
            return var

    def _vec_var(símismo, var, tx=False):
        """
        Devuelve un vector de los valores de un variable.

        Parameters
        ----------
        var: str
            El nombre del variable.
        tx: bool
            Si quieres formato de texto o numérico.

        Returns
        -------
        np.ndarray
        """
        raise NotImplementedError

    @staticmethod
    def _filtrar_lugares(vals, criteria):
        if criteria is None:
            return vals
        criteria = [criteria] if isinstance(criteria, str) else criteria
        return vals.where(vals[_('lugar')].isin(criteria), drop=True)

    @staticmethod
    def _filtrar_fechas(vals, criteria):
        if criteria is None:
            return vals
        criteria = pd.to_datetime(criteria)
        if isinstance(criteria, pd.Timestamp):
            criteria = criteria.to_datetime64()
        fechas = vals[_('fecha')]
        if isinstance(criteria, tuple) and len(criteria) == 2:
            cond = np.logical_and(np.less_equal(fechas, criteria[1]), np.greater_equal(fechas, criteria[0]))
        else:
            cond = fechas.isin(criteria)
        return vals.where(cond, drop=True)

    def __str__(símismo):
        return símismo.nombre


class FuenteCSV(Fuente):
    """
    Fuente para archivos ``.csv``.
    """
    def __init__(símismo, archivo, nombre=None, lugares=None, fechas=None, cód_vacío=None):
        """

        Parameters
        ----------
        archivo: str
            El archivo con los datos.
        nombre: str
            El nombre de la fuente.
        lugares: str or np.ndarray or list
            Los lugares que corresponden a los datos. Puede se nombre de una columna en el csv, el nombre de un
            lugar de cual vienen todos los datos, o una lista de los lugares.
        fechas: str or np.ndarray or list or datetime.datetime
            Las fechas de los datos.
        cód_vacío:
            Código para identificar variables que faltan. ``NA`` y ``NaN`` ya están reconocidos.
        """
        nombre = nombre or os.path.splitext(os.path.split(archivo)[1])[0]
        símismo.archivo = archivo
        símismo.codif = detectar_codif(archivo, máx_líneas=1)

        cód_vacío = cód_vacío or ['na', 'NA', 'NaN', 'nan', 'NAN', '']
        símismo.cód_vacío = [cód_vacío] if isinstance(cód_vacío, (int, float, str)) else cód_vacío

        super().__init__(nombre, variables=símismo.obt_vars(), lugares=lugares, fechas=fechas)

    def obt_vars(símismo):
        with open(símismo.archivo, encoding=símismo.codif) as d:
            lector = csv.reader(d)

            nombres_cols = next(lector)

        return nombres_cols

    def _vec_var(símismo, var, tx=False):
        l_datos = []
        with open(símismo.archivo, encoding=símismo.codif) as d:
            lector = csv.DictReader(d)

            for n_f, f in enumerate(lector):
                val = f[var].strip()
                if not tx:
                    if val in símismo.cód_vacío:
                        val = np.nan
                    else:
                        try:
                            val = எ(val)
                        except ValueError:
                            val = float(val)
                l_datos.append(val)

        return np.array(l_datos)


class FuenteDic(Fuente):
    """
    Fuente de datos en forma de diccionario.
    """

    def __init__(símismo, dic, nombre, lugares=None, fechas=None):
        """

        Parameters
        ----------
        dic: dict
            El diccionario con los datos.
        nombre: str
            El nombre de la fuente.
        lugares: str or np.ndarray or list
            Los lugares que corresponden a los datos. Puede se nombre de una llave en el dictionario, el nombre de un
            lugar de cual vienen todos los datos, o una lista de los lugares.
        fechas: str or np.ndarray or list or datetime.datetime
            Las fechas de los datos.
        """
        símismo.dic = dic
        super().__init__(nombre, variables=list(símismo.dic), lugares=lugares, fechas=fechas)

    def _vec_var(símismo, var, tx=False):
        return np.array(símismo.dic[var])


class FuenteVarXarray(Fuente):
    """
    Fuente para datos en formato de ``DataArray`` de ``xarray``.
    """

    def __init__(símismo, obj, nombre, lugares=None, fechas=None):
        """

        Parameters
        ----------
        obj: xarray.DataArray
            Los datos
        nombre: str
            El nombre de la fuente.
        lugares: str or np.ndarray or list
            Los lugares que corresponden a los datos. Puede se nombre de una columna en el ``DataArray``, el nombre de
            un lugar de cual vienen todos los datos, o una lista de los lugares.
        fechas: str or np.ndarray or list or datetime.datetime
            Las fechas de los datos.
        """
        símismo.obj = obj
        super().__init__(nombre, variables=[símismo.obj.name], lugares=lugares, fechas=fechas)

    def _vec_var(símismo, var, tx=False):
        return símismo.obj


class FuenteBaseXarray(Fuente):
    """
    Fuente para datos en formato de ``Dataset`` de ``xarray``.
    """
    def __init__(símismo, obj, nombre, lugares=None, fechas=None):
        """

        Parameters
        ----------
        obj: xarray.Dataset
            Los datos
        nombre: str
            El nombre de la fuente.
        lugares: str or np.ndarray or list
            Los lugares que corresponden a los datos. Puede se nombre de una columna en el ``Dataset``, el nombre de un
            lugar de cual vienen todos los datos, o una lista de los lugares.
        fechas: str or np.ndarray or list or datetime.datetime
            Las fechas de los datos.
        """
        símismo.obj = obj
        super().__init__(nombre, variables=list(símismo.obj.data_vars), lugares=lugares, fechas=fechas)

    def _vec_var(símismo, var, tx=False):
        return símismo.obj[var]


class FuentePandas(Fuente):
    """
    Fuente para datos en formato de ``DataFrame`` de ``xarray``.
    """
    def __init__(símismo, obj, nombre, lugares=None, fechas=None):
        """

        Parameters
        ----------
        obj: pd.DataFrame
            Los datos
        nombre: str
            El nombre de la fuente.
        lugares: str or np.ndarray or list
            Los lugares que corresponden a los datos. Puede se nombre de una columna en el ``Dataset``, el nombre de un
            lugar de cual vienen todos los datos, o una lista de los lugares.
        fechas: str or np.ndarray or list or datetime.datetime
            Las fechas de los datos.
        """
        símismo.obj = obj
        super().__init__(nombre, variables=list(símismo.obj), lugares=lugares, fechas=fechas)

    def _vec_var(símismo, var, tx=False):
        return símismo.obj[var]



import numpy as np
from dateutil.relativedelta import relativedelta as deltarelativo

from tinamit.envolt.bf._deter import ModeloDeterminado, VariablesModDeter, VarPasoDeter
from tinamit.tiempo.tiempo import a_unid_tnmt, a_unid_ft


class ModeloBloques(ModeloDeterminado):

    def __init__(símismo, variables, nombre='bf'):
        """

        Parameters
        ----------
        variables: VariablesModBloques
        nombre: str
        """

        símismo.tmñ_bloques = variables.tmñ_bloques
        super().__init__(tmñ_ciclo=np.sum(símismo.tmñ_bloques), variables=variables, nombre=nombre)

    def _act_vals_clima(símismo, f_0, f_1, vars_clima=None):

        vars_clima = vars_clima or símismo.vars_clima

        # Actualizar datos de clima
        p = símismo.paso_en_ciclo

        if símismo.corrida.clima and vars_clima and p == (símismo.tmñ_ciclo - 1):
            t = símismo.corrida.t
            f_inic = t.fecha()

            un_día = deltarelativo(days=1)

            vars_clim_bloque = {
                vr: d for vr, d in vars_clima.items() if isinstance(símismo.variables[vr], VarBloque)
            }
            vars_otros = {vr: d for vr, d in vars_clima.items() if vr not in vars_clim_bloque}

            super()._act_vals_clima(f_0=f_0, f_1=f_1, vars_clima=vars_otros)

            base_t, factor = a_unid_tnmt(símismo.unidad_tiempo())

            for i, b in enumerate(símismo.tmñ_bloques):

                f_final = f_inic + deltarelativo(**{a_unid_ft[base_t]: factor * b}) - un_día

                # Calcular los datos
                datos = símismo.corrida.clima.combin_datos(
                    vars_clima=vars_clim_bloque, f_inic=f_inic, f_final=f_final
                )

                # Aplicar los valores de variables calculados
                for var, datos_vrs in datos.items():
                    # Guardar el valor para esta estación
                    símismo.variables[var].poner_vals_paso(datos_vrs, paso=i)

                f_inic = f_final + un_día

    def unidad_tiempo(símismo):
        raise NotImplementedError

    def avanzar_modelo(símismo, n_ciclos):
        raise NotImplementedError


class VariablesModBloques(VariablesModDeter):

    def __init__(símismo, variables, tmñ_bloques):
        símismo.tmñ_bloques = tmñ_bloques
        símismo.tmñ_bloques_cum = np.cumsum(tmñ_bloques)
        símismo.n_bloques = len(tmñ_bloques)
        super().__init__(variables)

    def act_paso(símismo, paso):
        b = símismo.bloque(paso)
        for v in símismo.vars_paso():
            if isinstance(v, VarBloque):
                v.act_paso(b)
            else:
                v.act_paso(paso)

    def bloque(símismo, paso):
        return next(i for i, s in enumerate(símismo.tmñ_bloques_cum) if paso < s)


class VarBloque(VarPasoDeter):

    def __init__(símismo, nombre, unid, ingr, egr, tmñ_bloques, inic=0, líms=None, info=''):
        super().__init__(nombre, unid, ingr, egr, tmñ_ciclo=len(tmñ_bloques), inic=inic, líms=líms, info=info)



from ._auto import gen_bf
from ._bloq import VarBloque, VariablesModBloques, ModeloBloques
from ._deter import ModeloDeterminado, VariablesModDeter, VarPasoDeter
from ._envolt import ModeloBF
from ._impac import VariablesModImpaciente, ModeloImpaciente
from ._indeter import ModeloIndeterminado, VarPasoIndeter, VariablesModIndeterminado



import os
import unittest

import numpy as np
from numpy import testing as npt

from pruebas.recursos.bf.prueba_bf import PruebaBF
from pruebas.recursos.bf.variantes import EjDeterminado, EjBloques, EjIndeterminado
from tinamit.envolt.bf import gen_bf, ModeloBF

dir_act = os.path.split(__file__)[0]
arch_bf = os.path.join(dir_act, 'recursos/bf/prueba_mod.py')


class TestModeloDeter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.res = EjDeterminado(tmñ_ciclo=5).simular(50)

    def test_paso(símismo):
        npt.assert_equal(símismo.res['paso'].vals.values.flatten(), np.arange(51))

    def test_ciclo(símismo):
        npt.assert_equal(símismo.res['ciclo'].vals.values.flatten(), np.concatenate(([0], np.repeat(np.arange(1, 11), 5))))

    def test_i_en_ciclo(símismo):
        npt.assert_equal(símismo.res['i_en_ciclo'].vals.values.flatten(), np.concatenate(([4], np.tile(np.arange(5), 10))))


class TestModeloIndeter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.res = EjIndeterminado(rango_n=(3, 7)).simular(50)

    def test_paso(símismo):
        npt.assert_equal(símismo.res['paso'].vals.values.flatten(), np.arange(51))

    def test_ciclo(símismo):
        ciclo = símismo.res['ciclo'].vals.values.flatten()
        npt.assert_equal(np.unique(ciclo), np.arange(len(np.unique(ciclo))))
        símismo.assertTrue(all(np.sum(ciclo == c) <= 7 for c in np.unique(ciclo)))


class TestModeloBloques(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.res = EjBloques(tmñ_bloques=[3, 4, 5]).simular(36)

    def test_paso(símismo):
        npt.assert_equal(símismo.res['paso'].vals.values.flatten(), np.arange(37))

    def test_ciclo(símismo):
        npt.assert_equal(
            símismo.res['ciclo'].vals.values.flatten(), np.concatenate(([0], np.repeat(np.arange(1, 4), 12)))
        )

    def test_bloque(símismo):
        npt.assert_equal(
            símismo.res['bloque'].vals.values.flatten(),
            np.concatenate(([2], np.tile(np.repeat([0, 1, 2], [3, 4, 5]), 3)))
        )

    def test_i_en_ciclo(símismo):
        npt.assert_equal(
            símismo.res['i_en_ciclo'].vals.values.flatten(), np.concatenate(([11], np.tile(np.arange(12), 3)))
        )


class TestGenAuto(unittest.TestCase):
    def test_instancia(símismo):
        mod = gen_bf(PruebaBF())
        símismo.assertIsInstance(mod, ModeloBF)

    def test_clase(símismo):
        mod = gen_bf(PruebaBF)
        símismo.assertIsInstance(mod, ModeloBF)

    def test_archivo_instancia(símismo):
        mod = gen_bf(símismo._obt_recurso('arch_instancia.py'))
        símismo.assertIsInstance(mod, ModeloBF)

    def test_archivo_clase(símismo):
        mod = gen_bf(símismo._obt_recurso('arch_clase.py'))
        símismo.assertIsInstance(mod, ModeloBF)

    def test_archivo_instancia_no_identificada(símismo):
        mod = gen_bf(símismo._obt_recurso('arch_instancia_no_identificada.py'))
        símismo.assertIsInstance(mod, ModeloBF)

    def test_archivo_clase_no_identificada(símismo):
        mod = gen_bf(símismo._obt_recurso('arch_clase_no_identificada.py'))
        símismo.assertIsInstance(mod, ModeloBF)

    def test_archivo_múltiples(símismo):
        mod = gen_bf(símismo._obt_recurso('arch_múltiples.py'))
        símismo.assertIsInstance(mod, ModeloBF)

    def test_archivo_múltiples_no_identificada(símismo):
        mod = gen_bf(símismo._obt_recurso('arch_múltiples_no_identificada.py'))
        símismo.assertIsInstance(mod, ModeloBF)

    def test_archivo_vacío(símismo):
        with símismo.assertRaises(AttributeError):
            gen_bf(símismo._obt_recurso('arch_vacío.py'))

    def test_tipo_no_válido(símismo):
        with símismo.assertRaises(TypeError):
            gen_bf(123)

    def test_archivo_no_existe(símismo):
        with símismo.assertRaises(FileNotFoundError):
            gen_bf('નમસ્તે')

    def test_archivo_no_python(símismo):
        with símismo.assertRaises(ValueError):
            gen_bf(símismo._obt_recurso('no_python.txt'))

    @staticmethod
    def _obt_recurso(archivo):
        return os.path.join('recursos/bf/_prbs_arch', archivo)



import numpy as np

from tinamit.envolt.bf import ModeloBloques, ModeloDeterminado, ModeloIndeterminado, VariablesModDeter, \
    VariablesModIndeterminado, VarBloque, VarPasoDeter, VariablesModBloques, VarPasoIndeter
from tinamit.mod import Variable


class EjDeterminado(ModeloDeterminado):

    def __init__(símismo, tmñ_ciclo, unid_tiempo='días'):
        símismo.unid_tiempo = unid_tiempo

        super().__init__(tmñ_ciclo=tmñ_ciclo, variables=símismo._gen_vars(tmñ_ciclo))

    @staticmethod
    def _gen_vars(tmñ_ciclo):
        return VariablesModDeter([
            Variable('ciclo', unid=None, ingr=False, egr=True, líms=(0, None)),
            VarPasoDeter('paso', unid=None, ingr=False, egr=True, tmñ_ciclo=tmñ_ciclo, líms=(0, None)),
            VarPasoDeter(
                'i_en_ciclo', unid=None, ingr=False, egr=True, inic=tmñ_ciclo - 1, tmñ_ciclo=tmñ_ciclo, líms=(0, None)
            ),
            VarPasoDeter('ingreso_paso', unid=None, ingr=True, egr=False, tmñ_ciclo=tmñ_ciclo, líms=(0, None)),
            Variable('ingreso_ciclo', unid=None, ingr=True, egr=False, líms=(None, None))
        ])

    def unidad_tiempo(símismo):
        return símismo.unid_tiempo

    def avanzar_modelo(símismo, n_ciclos):
        símismo.variables['ciclo'] += n_ciclos
        paso = símismo.variables['paso'].obt_val()
        símismo.variables['paso'].poner_vals_paso(np.arange(paso + 1, paso + símismo.tmñ_ciclo + 1))
        símismo.variables['i_en_ciclo'].poner_vals_paso(np.arange(símismo.tmñ_ciclo))


class EjBloques(ModeloBloques):

    def __init__(símismo, tmñ_bloques, unid_tiempo='días'):
        símismo.unid_tiempo = unid_tiempo

        super().__init__(variables=símismo._gen_vars(tmñ_bloques))

    @staticmethod
    def _gen_vars(tmñ_bloques):
        # noinspection PyTypeChecker
        tmñ_ciclo = np.sum(tmñ_bloques)  # type: int
        return VariablesModBloques([
            Variable('ciclo', unid=None, ingr=False, egr=True, líms=(0, None)),
            VarBloque(
                'bloque', unid=None, ingr=False, egr=True, tmñ_bloques=tmñ_bloques, inic=len(tmñ_bloques) - 1,
                líms=(0, None)
            ),
            VarPasoDeter('paso', unid=None, ingr=False, egr=True, tmñ_ciclo=tmñ_ciclo, líms=(0, None)),
            VarPasoDeter(
                'i_en_ciclo', unid=None, ingr=False, egr=True, tmñ_ciclo=tmñ_ciclo, inic=tmñ_ciclo - 1, líms=(0, None),
            ),
            Variable('ingreso_ciclo', unid=None, ingr=True, egr=False, líms=(0, None)),
            VarPasoDeter('ingreso_paso', unid=None, ingr=True, egr=False, tmñ_ciclo=tmñ_ciclo, líms=(0, None)),
            VarBloque('ingreso_bloque', unid=None, ingr=True, egr=False, tmñ_bloques=tmñ_bloques, líms=(0, None))
        ], tmñ_bloques=tmñ_bloques)

    def unidad_tiempo(símismo):
        return símismo.unid_tiempo

    def avanzar_modelo(símismo, n_ciclos):
        símismo.variables['ciclo'] += n_ciclos
        paso = símismo.variables['paso'].obt_val()
        símismo.variables['paso'].poner_vals_paso(np.arange(paso + 1, paso + símismo.tmñ_ciclo + 1))
        símismo.variables['i_en_ciclo'].poner_vals_paso(np.arange(símismo.tmñ_ciclo))
        símismo.variables['bloque'].poner_vals_paso(np.arange(len(símismo.variables.tmñ_bloques)))


class EjIndeterminado(ModeloIndeterminado):

    def __init__(símismo, rango_n, unid_tiempo='días'):
        símismo.rango_n = rango_n
        símismo.unid_tiempo = unid_tiempo

        super().__init__(variables=símismo._gen_vars())

    @staticmethod
    def _gen_vars():
        return VariablesModIndeterminado([
            Variable('ciclo', unid=None, ingr=False, egr=True, líms=(0, None)),
            VarPasoIndeter('paso', unid=None, ingr=False, egr=True, líms=(0, None)),
            Variable('ingreso', unid=None, ingr=True, egr=False, líms=(0, None))
        ])

    def unidad_tiempo(símismo):
        return símismo.unid_tiempo

    def mandar_modelo(símismo):
        n = np.random.randint(*símismo.rango_n)
        símismo.variables['ciclo'] += 1
        paso = símismo.variables['paso'].obt_val()
        símismo.variables['paso'].poner_vals_paso(np.arange(paso + 1, n + paso + 1))
        return n



from warnings import warn as avisar

import numpy as np
from scipy.optimize import minimize

from tinamit.calibs._utils import eval_funcs
from tinamit.calibs.ec import CalibradorEc
from tinamit.config import _
from tinamit.geog.región import Lugar


class CalibradorEcOpt(CalibradorEc):
    """
    Calibrador de ecuaciones con algoritmo de optimización.
    """

    def __init__(símismo, ec, paráms, nombre=None, dialecto='tinamït'):

        super().__init__(ec, paráms, nombre=nombre, dialecto=dialecto)

        símismo.f_python = símismo.ec.a_python(paráms=símismo.paráms)

    def calibrar(símismo, bd, lugar=None, líms_paráms=None, ops=None, corresp_vars=None, ord_niveles=None):
        """
        Efectua una calibración para cada lugar en ``Lugar`` según los datos en ``bd``.

        Parameters
        ----------
        bd: BD
            La base de datos con observaciones para los variables en la ecuación.
        lugar: Lugar
            El lugar cuyos sublugares hay que calibrar; si es ``None`` se calibrará la ecuación con todos
            los datos en ``bd`` sin tener su lugar en cuenta.
        líms_paráms: list
            Límites teoréticos para los parámetros.
        ops: dict
            Opciones que se pasarán directamente a la función de calibración.
        corresp_vars: dict
            Diccionario de correspondencia entre los nombres de los variables en ``bd`` y sus nombres en la ecuación.
        ord_niveles: list
            Desambiguación del orden de niveles.

        Returns
        -------
        dict
            Diccionario con las calibraciones de cada lugar.
        """

        ops = ops or {}

        líms_paráms = símismo._gen_líms_paráms(líms_paráms)

        # Leer los variables de la ecuación.
        vars_x, var_y = símismo._extraer_vars()

        # Todas las observaciones
        obs = símismo._obt_datos(bd, vars_interés=vars_x + [var_y], corresp_vars=corresp_vars)

        # Calibrar según la situación
        if lugar is None:
            # Si no hay lugares, optimizar de una vez con todas las observaciones.
            return _optimizar(símismo.f_python, líms_paráms=líms_paráms, obs_x=obs[vars_x], obs_y=obs[var_y], **ops)

        # Si hay lugares...

        # Una función recursiva para calibrar según la jerarquía
        def _calibrar_jerárchico_manual(lug, clbs):
            """
            Una función recursiva que permite calibrar en una jerarquía, tomando optimizaciones de niveles más
            altos si no consigue datos para niveles más bajos.

            Parameters
            ----------
            lug: Lugar
                El lugar en cual calibrar.
            clbs: dict
                El diccionario de los resultados de la calibración. (Parámetro recursivo.)
            """

            # Sino, tomar los datos de esta región únicamente.
            obs_lg = obs.where(obs['lugar'].isin([x.cód for x in lug.lugares()]), drop=True)

            # Intentar sacar información del nivel superior en la jerarquía

            if obs_lg.sizes['n']:

                def _buscar_inic_pariente(hj):
                    pariente = lugar.pariente(hj, ord_niveles=ord_niveles)  # El nivel inmediatamente superior
                    if pariente is None:
                        return
                    elif pariente.cód in clbs:
                        # Tomar la calibración superior como punto inicial para facilitar la búsqueda
                        return [clbs[pariente.cód][p]['cumbre'] for p in símismo.paráms]
                    else:
                        return _buscar_inic_pariente(pariente)

                inic = _buscar_inic_pariente(lug)

                resultados[lug.cód] = _optimizar(
                    símismo.f_python, líms_paráms=líms_paráms,
                    obs_x=obs_lg[vars_x], obs_y=obs_lg[var_y], inic=inic, **ops
                )
            for sub in lug.hijos_inmediatos(ord_niveles):
                _calibrar_jerárchico_manual(lug=sub, clbs=clbs)

        # Calibrar para cada lugar
        resultados = {}
        _calibrar_jerárchico_manual(lug=lugar, clbs=resultados)

        return resultados


def _optimizar(func, líms_paráms, obs_x, obs_y, inic=None, **ops):
    """
    Optimiza una función basándose en observaciones.

    Parameters
    ----------
    func: Callable
        La función para optimizar.
    líms_paráms: dict[str, tuple]
        Un diccionario de los parámetros y de sus límites.
    obs_x: pd.DataFrame
        Las observaciones de los variables x.
    obs_y: pd.Series | np.array
        Las observaciones correspondientes del variable y.
    inic: list | np.array
        Los valores iniciales para la optimización.
    ops: dict
        Opciones para pasar a la función de optimización.

    Returns
    -------
    dict
        Los parámetros optimizados.
    """

    # Leer el método de ajusto.
    try:
        med_ajusto = ops.pop('med_ajusto')
    except KeyError:
        med_ajusto = 'rcep'

    # La lista de parámetros de interés.
    paráms = list(líms_paráms)

    # Crear la función objetiva que minimizaremos (SciPy solamente puede minimizar).
    def f(prm):
        """
        Una función objetiva que SciPy puede minimizar.

        Parameters
        ----------
        prm: list | np.ndarray
            Los parámetros para calibrar

        Returns
        -------
        np.ndarray
            El ajusto del modelo con los parámetros actuales.

        """

        f_ajusto = eval_funcs[med_ajusto.lower()]

        # Devolver el ajusto de la función con los parámetros actuales.
        return -f_ajusto(func(prm, obs_x), obs_y, f=None)  # SpotPy maximiza pero nosotr@s aquí minimizamos

    # Generar los estimos iniciales, si necesario
    if inic is not None:
        # No es necesario
        x0 = inic
    else:
        # Sí es necesario
        x0 = []
        for p in paráms:
            # Para cada parámetro...
            lp = líms_paráms[p]  # Sus límites

            # Calcular un punto razonable para empezar la búsqueda
            if lp[0] is None:
                if lp[1] is None:
                    # El caso (-inf, +inf): empezaremos en 0
                    x0.append(0)
                else:
                    # El caso (-inf, R]: empezaremos en R
                    x0.append(lp[1])
            else:
                if lp[1] is None:
                    # El caso [R, +inf): empezaremos en R
                    x0.append(lp[0])
                else:
                    # El caso [R1, R2]: empezaremos en el promedio de R1 y R2
                    x0.append((lp[0] + lp[1]) / 2)

    # Convertir a matriz NumPy
    x0 = np.array(x0)

    # Optimizar
    opt = minimize(f, x0=x0, bounds=[líms_paráms[p] for p in paráms], **ops)

    # Avisar si la optimización no funcionó tan bien como lo esperamos.
    if not opt.success:
        avisar(_('Es posible que haya un error de optimización. Mejor le eches un vistazo a los resultados.'))

    # Devolver los resultados en el formato correcto.
    return {p: {'cumbre': opt.x[i]} for i, p in enumerate(paráms)}



import unittest

from tinamit.envolt.bf.sahysmod import ModeloSAHYSMOD
from tinamit.mod.prbs import verificar_leer_ingr, verificar_leer_egr, verificar_simul

_envolts = [ModeloSAHYSMOD]


class TestEnvolturasExtern(unittest.TestCase):

    def test_leer_ingreso(símismo):
        for cls in _envolts:
            with símismo.subTest(envlt=cls.__name__):
                verificar_leer_ingr(símismo, cls)

    def test_leer_egreso(símismo):
        for cls in _envolts:
            with símismo.subTest(envlt=cls.__name__):
                verificar_leer_egr(símismo, cls)

    def test_correr_modelo(símismo):
        for cls in _envolts:
            with símismo.subTest(envlt=cls.__name__):
                verificar_simul(símismo, cls)



from warnings import warn as avisar

from tinamit.config import _
from tinamit.unids.conv import convertir


class ConexionesVars(object):
    def __init__(símismo):
        símismo._conexiones = []

    def agregar(símismo, conex):
        if any(conex.var_recip is c.var_recip for c in símismo):
            raise ValueError(_('El variable "{}" ya está conectado como variable recipiente.').format(conex.var_recip))

        # Si esta conexíon depende de otra, asegurarse que la otra se actualice antes de ésta
        i = next((i + 1 for i, c in enumerate(símismo._conexiones) if (c.var_recip is conex.var_fuente)), 0)
        símismo._conexiones.insert(i, conex)

    def quitar(símismo, var_fuente, modelo_fuente, modelo_recip, var_recip):
        for c in símismo._buscar(var_fuente, modelo_fuente, modelo_recip, var_recip):
            símismo._conexiones.remove(c)

    def _buscar(símismo, var_fuente, modelo_fuente, modelo_recip, var_recip):
        return [
            c for c in símismo
            if (c.var_fuente is var_fuente) and (c.modelo_fuente is modelo_fuente) and (
                    modelo_recip is None or c.modelo_recip is modelo_recip) and (
                       var_recip is None or c.var_recip is var_recip)
        ]

    def __iter__(símismo):
        for c in símismo._conexiones:
            yield c


class Conex(object):
    def __init__(símismo, var_fuente, modelo_fuente, var_recip, modelo_recip, conv):

        if modelo_recip is modelo_fuente:
            raise ValueError(_('Los modelos de variables conectados deben ser distintos.'))

        símismo.var_fuente = var_fuente
        símismo.modelo_fuente = modelo_fuente
        símismo.var_recip = var_recip
        símismo.modelo_recip = modelo_recip

        símismo._verificar()

        símismo.conv = _resolv_conv(var_fuente.unid, var_recip.unid, conv=conv)

    def _verificar(símismo):
        v_f = símismo.var_fuente
        v_r = símismo.var_recip

        if v_f.dims != v_r.dims:
            raise ValueError(
                _('Dimensiones incompatibles entre variables "{}" {} y "{}" {}.').format(v_f, v_f.dims, v_r, v_r.dims)
            )

        if not _líms_compat(símismo.var_fuente.líms, símismo.var_recip.líms):
            avisar(_('Límites potencialmente incompatibles entre variables "{}" y "{}".').format(
                símismo.var_fuente, símismo.var_recip)
            )


def _resolv_conv(unid_1, unid_2, conv):
    if conv is None:
        # Si no se especificó factor de conversión...

        mensaje = _('No se pudo identificar una conversión automática entre {} y {}. '
                    'Se está suponiendo un factor de conversión de 1.').format(unid_1, unid_2)
        if unid_1 is None or unid_2 is None:
            conv = 1
            if not (unid_1 is unid_2 is None):
                avisar(mensaje)

        else:
            # Intentar hacer una conversión automática.
            try:
                conv = convertir(de=unid_1, a=unid_2)
            except ValueError:
                # Si eso no funcionó, suponer una conversión de 1.
                conv = 1
                avisar(mensaje)

    return conv


def _líms_compat(líms_1, líms_2):
    return (líms_1[0] >= líms_2[0]) and (líms_1[1] <= líms_2[1])



import os
import unittest

import numpy as np
import numpy.testing as npt
import pandas as pd
import scipy.stats as estad
from tinamit.calibs.ec import CalibradorEcOpt, CalibradorEcBayes
from tinamit.datos.bd import BD
from tinamit.datos.fuente import FuenteDic
from tinamit.geog.región import gen_lugares
from tinamit.ejemplos import obt_ejemplo

try:
    import pymc3 as pm
    import theano as thn
except ImportError:
    pm = thn = None

dir_act = os.path.split(__file__)[0]
arch_csv_geog = obt_ejemplo('geog_guate/geog_guate.csv')
arch_mds = os.path.join(dir_act, 'recursos/mds/prueba_para_calib.mdl')

calibradores = {'opt': CalibradorEcOpt, 'bayes': CalibradorEcBayes}
if pm is None or thn.configdefaults.rc != 0:  # saltar si no hay pymc3 o si no hay compilador c
    calibradores.pop('bayes')


class TestCalibrador(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        n_obs = 50
        datos_x = np.random.rand(n_obs)

        ec = 'y = a*x + b'
        cls.paráms = {'a': 2.4, 'b': -5}

        cls.clbrds = {ll: v(ec=ec, paráms=list(cls.paráms)) for ll, v in calibradores.items()}

        datos_y = cls.paráms['a'] * datos_x + cls.paráms['b'] + np.random.normal(0, 0.1, n_obs)
        cls.bd_datos = BD(
            fuentes=FuenteDic({'y': datos_y, 'x': datos_x, 'f': np.arange(n_obs)}, 'Datos generados', fechas='f')
        )

    def test_calibración_sencilla(símismo):
        líms = {
            'sin_líms': None,
            'un_lím': {'a': (0, None), 'b': (None, 0)},
            'dos_líms': {'a': (0, 10), 'b': (-10, -1)}
        }
        for nmbr, clbrd in símismo.clbrds.items():
            for lm in líms:
                with símismo.subTest(método=nmbr, líms=lm):
                    calibs = clbrd.calibrar(líms_paráms=líms[lm], bd=símismo.bd_datos)

                    est = [calibs[p]['cumbre'] for p in símismo.paráms]
                    val = list(símismo.paráms.values())
                    npt.assert_allclose(est, val, rtol=0.1)

    def test_calibrador_sin_var_y(símismo):
        with símismo.assertRaises(ValueError):
            CalibradorEcOpt(ec='a*x+b', paráms=list(símismo.paráms))


class TestCalibGeog(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.paráms = prms = {
            'a': {'701': 3.4, '708': 3, '1001': 10},
            'b': {'701': -1.5, '708': -1.1, '1001': -3}
        }  # Nota: ¡a en (0, +inf) y b en (-inf, +inf) según los límites en el modelo externo!
        cls.ec = 'y = a*x + b'
        n_obs = {'701': 500, '708': 50, '1001': 500}
        datos_x = {lg: np.random.rand(n) for lg, n in n_obs.items()}
        datos_y = {
            lg: datos_x[lg] * prms['a'][lg] + prms['b'][lg] + np.random.normal(0, 0.1, n_obs[lg])
            for lg in n_obs
        }  # y = a*x + b
        lugares = [x for ll, v in datos_x.items() for x in [ll] * v.size]
        x = [i for v in datos_x.values() for i in v]
        y = [i for v in datos_y.values() for i in v]

        cls.clbrds = {ll: v(ec=cls.ec, paráms=['a', 'b']) for ll, v in calibradores.items()}

        fchs = pd.date_range(0, periods=len(x))
        cls.bd = BD(FuenteDic(
            {'lugar': lugares, 'x': x, 'y': y, 'f': fchs}, 'Datos geográficos', lugares='lugar', fechas='f'
        ))

        cls.lugar = gen_lugares(arch_csv_geog, nivel_base='País', nombre='Iximulew')

    def test_calibración_geog_con_escalas(símismo):
        """
        Calibramos una geografía con distintas escalas (municipios y departamentos).
        """
        líms_paráms = {'a': (0, 50)}
        for m in calibradores:
            with símismo.subTest(método=m):
                clbrd = calibradores[m](símismo.ec, paráms=['a', 'b'])
                calibs = clbrd.calibrar(símismo.bd, lugar=símismo.lugar, líms_paráms=líms_paráms)

                val = [símismo.paráms[p][lg] for p in símismo.paráms for lg in símismo.paráms[p]]
                if m == 'opt':
                    est = [calibs[lg][p]['cumbre'] for p in símismo.paráms for lg in símismo.paráms[p]]
                    npt.assert_allclose(val, est, rtol=0.2)
                else:
                    est = [calibs[lg][p]['dist'] for p in símismo.paráms for lg in símismo.paráms[p]]
                    símismo._verificar_aprox_bayes(val, est)

    def test_calibración_bayes_sin_mod_jerárquíco(símismo):
        """
        Calibramos en una geografía sin modelo jerárquico (región por región sin información a priori).
        """

        if 'bayes' in calibradores:
            calibs = símismo.clbrds['bayes'].calibrar(símismo.bd, lugar=símismo.lugar, jerárquico=False)
            val = [símismo.paráms[lg][p] for lg in símismo.paráms for p in símismo.paráms[lg]]
            est = [calibs[p][lg]['dist'] for lg in símismo.paráms for p in símismo.paráms[lg]]

            símismo._verificar_aprox_bayes(val, est)

    @staticmethod
    def _verificar_aprox_bayes(val, est, intvl=99):
        npt.assert_allclose([estad.percentileofscore(e, v) for e, v in zip(est, val)], 50, atol=intvl / 2)



"""
Código modificado de Jérôme Boisvert-Chouinard bajo la licencia MIT.
Code modified from Jérôme Boisvert-Chouinard under the MIT license.

The MIT License (MIT)

Copyright (c) 2014 Jerome Boisvert-Chouinard

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# !/bin/python

# _ie.py

import os
import sys

import numpy as np
from pkg_resources import resource_filename

from tinamit.envolt.bf.sahysmod._arch_ingr import _anlzd_sntct

PLANTILLACSV = resource_filename(__name__, 'rcrs/sahysmod.csv.tmpl')
PLANTILLAINP = resource_filename(__name__, 'rcrs/sahysmod.inp.tmpl')

parámsParaTransponer = [
    'A',
    'B',
    'KCA',
    'KCB',
    'PP',
    'EPA',
    'EPB',
    'EPU',
    'SIU',
    'SOU',
    'SOA',
    'SOB',
    'GW',
    'FRD',
    'LC',
    'CI',
    'IAA',
    'IAB',
    'GU',
    'FW',
    'FSA',
    'FSB',
    'FSU',
    'HS'
]

paráms_enteros = [
    'N_IN',
    'N_EX',
    'KIE_IN',
    'KIE_EX',
    'KLC',
    'FROM_N',
    'TO_N1',
    'TO_N2',
    'TO_N3',
    'TO_N4',
    'KD',
    'KF',
    'KR',
    'KRF',
]


def transponer_paráms(dic_paráms):
    for parám in parámsParaTransponer:
        matr = dic_paráms[parám]  # type: np.ndarray
        matr.transpose()


def leer_info_dic_paráms(archivo_fnt):
    """
    Convierte un fuente de ingresos SAHYSMOD en un diccionario de valores de variables.

    Parameters
    ----------
    archivo_fnt: str
        El fuente con los datos de ingreso, en formato .inp o .csv. 
        
    Returns
    -------
    dict
        El diccionario de los variables y sus valores.

    """

    # Escoger la plantilla apropiada.
    plantilla_fnt = PLANTILLACSV if archivo_fnt[-3:] == 'csv' else PLANTILLAINP

    # Leer el fuente de ingresos.
    dic_paráms = _anlzd_sntct.leer_archivo(archivo_fnt, plantilla_fnt, paráms_ent=paráms_enteros)
    transponer_paráms(dic_paráms)  # Transpose parameters that require it.

    n_en_s = []
    estación = []
    nn_in = int(dic_paráms['NN_IN'])
    for i in range(nn_in):
        n_en_s.append(('{}'.format(i), '', ''))
        estación.append(('1', '2', '3'))

    dic_paráms['NINETYNINE'] = '99'
    dic_paráms['ONE'] = '1'
    dic_paráms['FOURS'] = ['4'] * nn_in
    dic_paráms['N_IN_S'] = n_en_s
    dic_paráms['SEASON'] = estación
    dic_paráms['NNS'] = [str(i) for i in range(1, int(dic_paráms['NS']) + 1)]

    return dic_paráms


def escribir_desde_dic_paráms(dic_paráms, archivo_obj, csv=False):
    plantilla_obj = PLANTILLACSV if csv else PLANTILLAINP

    _anlzd_sntct.escribir_archivo(dic_paráms, archivo_obj, plantilla_obj, paráms_ent=paráms_enteros)


def central(archivo_fnt, archivo_obj):  # pragma: sin cobertura
    dic_paráms = leer_info_dic_paráms(archivo_fnt=archivo_fnt)

    escribir_desde_dic_paráms(dic_paráms=dic_paráms, archivo_obj=archivo_obj,
                              csv=archivo_obj[-3:] == 'csv')
    return 0


if __name__ == '__main__':
    # El código abajo se ejecutará si se llama este fuente desde la línea de comanda (o, por tanto, desde cualquier
    # otro lugar)

    # Leer los argumentos de la línea de comanda. (El primero, sys.argv[0], es simplemente la dirección de este
    # fuente y se ignorará. El segundo es el fuente desde cual hay que leer y el tercero él hacia cual hay que
    # escribir.)
    try:
        archivoFnt = os.path.join(os.getcwd(), sys.argv[1])
        archivoObj = os.path.join(os.getcwd(), sys.argv[2])
    except IndexError:
        # If arguments were not passed, use default file names. This is useful for debugging!
        archivoObj = os.path.join(os.getcwd(), "459anew1.csv")
        archivoFnt = os.path.join(os.getcwd(), "459anew1.inp")

    # Ahora, correr la función central.
    central(archivoFnt, archivoObj)



from .corrida import Corrida, OpsSimulGrupoCombin, OpsSimulGrupo
from .modelo import Modelo
from .res import ResultadosVar, ResultadosSimul, ResultadosGrupo
from .var import Variable
from .vars_mod import VariablesMod



import os
import unittest

import numpy as np
from tinamit.calibs.geog_mod import SimuladorGeog, CalibradorGeog, ValidadorGeog
from tinamit.calibs.mod import CalibradorModSpotPy
from tinamit.calibs.valid import ValidadorMod
from tinamit.datos.bd import BD
from tinamit.datos.fuente import FuenteDic
from tinamit.envolt.mds import gen_mds
from tinamit.ejemplos import obt_ejemplo

dir_act = os.path.split(__file__)[0]
arch_mds = obt_ejemplo('enfermedad/mod_enferm.mdl')
arch_csv_geog = obt_ejemplo('geog_guate/geog_guate.csv')

líms_paráms = {
    'taza de contacto': (0, 100),
    'taza de infección': (0, 0.02),
    'número inicial infectado': (0, 50),
    'taza de recuperación': (0, 0.1)
}


class TestCalibModelo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.paráms = {
            'taza de contacto': 81.25,
            'taza de infección': 0.007,
            'número inicial infectado': 22.5,
            'taza de recuperación': 0.0375
        }
        cls.mod = gen_mds(arch_mds)

        simul = cls.mod.simular(
            t=100,
            extern=cls.paráms,
            vars_interés=['Individuos Suceptibles', 'Individuos Infectados', 'Individuos Resistentes']
        )
        datos = {ll: v[:, 0] for ll, v in simul.a_dic().items()}  # Para hacer: dimensiones múltiples,
        cls.datos = FuenteDic(
            datos, 'Datos', fechas=np.arange(101)
        )

    def test_calibrar_validar(símismo):
        calibs = CalibradorModSpotPy(símismo.mod).calibrar(
            líms_paráms=líms_paráms,
            datos=símismo.datos,
            n_iter=50
        )
        valid = ValidadorMod(símismo.mod).validar(
            t=100, datos=símismo.datos, paráms={prm: trz['mejor'] for prm, trz in calibs.items()}
        )
        símismo.assertGreater(valid['Individuos Suceptibles']['ens'], 0.95)


class TestCalibModeloEspacial(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.paráms = {
            '708': {
                'taza de contacto': 81.25, 'taza de infección': 0.007, 'número inicial infectado': 22.5,
                'taza de recuperación': 0.0375
            },
            '1010': {
                'taza de contacto': 50, 'taza de infección': 0.005, 'número inicial infectado': 40,
                'taza de recuperación': 0.050
            }
        }
        cls.mod = gen_mds(arch_mds)
        simul = SimuladorGeog(cls.mod).simular(
            t=100, vals_geog=cls.paráms,
            vars_interés=['Individuos Suceptibles', 'Individuos Infectados', 'Individuos Resistentes']
        )

        # Para hacer: dimensiones múltiples,
        datos = {
            lg: {ll: v[:, 0] for ll, v in simul[lg].a_dic().items()} for lg in cls.paráms
        }

        cls.datos = BD([
            FuenteDic(datos[lg], 'Datos geográficos', lugares=lg, fechas=np.arange(101)) for lg in cls.paráms
        ])

    def test_calib_valid_espacial(símismo):
        calib = CalibradorGeog(símismo.mod).calibrar(t=100, datos=símismo.datos, líms_paráms=líms_paráms, n_iter=50)
        valid = ValidadorGeog(símismo.mod).validar(
            t=100, datos=símismo.datos,
            paráms={lg: {prm: trz['mejor'] for prm, trz in calib[lg].items()} for lg in símismo.paráms}
        )

        for lg in símismo.paráms:
            símismo.assertTrue(valid[lg]['Individuos Suceptibles']['ens'], 0.95)



import os
import random
import shutil
import tempfile
import unittest

import numpy as np

from pruebas.recursos.mod.prueba_mod import ModeloPrueba
from tinamit.ejemplos import obt_ejemplo
import tinamit.geog.mapa as mapa
from tinamit.geog.región import Nivel, Lugar, gen_lugares
from tinamit.mod import OpsSimulGrupo

dir_act = os.path.split(__file__)[0]
arch_csv_geog = obt_ejemplo('geog_guate/geog_guate.csv')
arch_frm_nombrada = obt_ejemplo('geog_guate/munis.shp')
arch_frm_numérica = obt_ejemplo('rechna_doab/polígonos.shp')
arch_frm_otra = obt_ejemplo('rechna_doab/canal.shp')


class TestRegión(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        muni = Nivel('Municipio')
        dept = Nivel('Departamento', subniveles=muni)
        terr = Nivel('Territorio', subniveles=muni)
        país = Nivel('País', subniveles=[dept, terr])
        muni1, muni2, muni3 = [Lugar('Muni%i' % i, nivel=muni, cód='M' + str(i)) for i in range(1, 4)]

        dept1 = Lugar('Dept1', nivel=dept, cód='D1', sub_lugares=[muni1, muni2])
        dept2 = Lugar('Dept2', nivel=dept, cód='D2', sub_lugares=[muni3])
        terr1 = Lugar('Terr1', nivel=terr, cód='T1', sub_lugares=[muni1])
        terr2 = Lugar('Terr2', nivel=terr, cód='T2', sub_lugares=[muni2])
        cls.guate = Lugar(
            'Guatemala', sub_lugares={muni1, muni2, muni3, dept1, dept2, terr1, terr2},
            nivel=país
        )

        cls.munis = {'1': muni1, '2': muni2, '3': muni3}
        cls.depts = {'1': dept1, '2': dept2}
        cls.terrs = {'1': terr1, '2': terr2}

    def test_obt_lugares(símismo):
        símismo.assertSetEqual(
            símismo.guate.lugares(),
            {símismo.guate, *símismo.munis.values(), *símismo.depts.values(), *símismo.terrs.values()}
        )

    def test_obt_lugares_con_nivel(símismo):
        símismo.assertSetEqual(símismo.guate.lugares(nivel='Municipio'), set(símismo.munis.values()))

    def test_obt_lugares_en(símismo):
        símismo.assertSetEqual(
            símismo.guate.lugares(en='D1', nivel='Municipio'),
            símismo.depts['1'].sub_lugares
        )

    def test_parientes(símismo):
        símismo.assertEqual(
            símismo.guate.pariente('M1', 'Departamento'), símismo.depts['1']
        )

    def test_hijos(símismo):
        símismo.assertSetEqual(
            {x.cód for x in símismo.guate.hijos_inmediatos(ord_niveles=['Territorio', 'Municipio', 'País'])},
            {x.cód for x in list(símismo.depts.values()) + list(símismo.terrs.values()) + [símismo.munis['3']]}
        )

    def test_obt_lugares_donde_no_hay(símismo):
        símismo.assertSetEqual(
            símismo.guate.lugares(en='D1', nivel='Territorio'),
            set()
        )

    def test_buscar_de_código(símismo):
        símismo.assertEqual(símismo.guate['M1'], símismo.munis['1'])

    def test_buscar_de_nombre(símismo):
        símismo.assertEqual(símismo.guate.buscar_nombre('Muni1'), símismo.munis['1'])

    def test_buscar_código_no_existe(símismo):
        with símismo.assertRaises(ValueError):
            símismo.guate.buscar_nombre('Hola, yo no existo.')


class TestGenAuto(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.lugar = gen_lugares(arch_csv_geog, nivel_base='País', nombre='Iximulew')

    def test_niveles(símismo):
        nivel = símismo.lugar.nivel
        símismo.assertEqual(nivel, 'País')
        símismo.assertSetEqual(set(str(n) for n in nivel.subniveles), {'Departamento', 'Territorio'})
        símismo.assertEqual(nivel['Departamento'].subniveles, 'Municipio')

    def test_lugares(símismo):
        símismo.assertEqual(símismo.lugar.buscar_nombre('SOLOLÁ', 'Departamento').cód, '7')
        símismo.assertEqual(símismo.lugar.buscar_nombre('SOLOLÁ', 'Municipio').cód, '701')


class TestMapa(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dir_ = tempfile.mkdtemp(prefix='prb_mapa_')

    def _verificar_dibujó(símismo, formas, nombre):
        arch = os.path.join(símismo.dir_, nombre)
        mapa.dibujar_mapa(formas, arch)
        símismo.assertTrue(os.path.isfile(arch))
        os.remove(arch)

    def test_dibujar_formas_estáticas(símismo):
        frms = [cls(arch_frm_otra) for cls in [mapa.Calle, mapa.Ciudad, mapa.Bosque, mapa.OtraForma, mapa.Agua]]
        símismo._verificar_dibujó(frms, 'mapa_estático.jpeg')

    def test_dibujar_forma_numérica(símismo):
        frm = mapa.FormaDinámicaNumérica(arch_frm_numérica, col_id='Id')
        frm.estab_valores(np.random.rand(*frm.valores.shape))
        símismo._verificar_dibujó(frm, 'mapa_matriz.jpeg')

    def test_dibujar_forma_numérica_sin_id(símismo):
        frm = mapa.FormaDinámicaNumérica(arch_frm_numérica)
        frm.estab_valores(np.random.rand(*frm.valores.shape))
        símismo._verificar_dibujó(frm, 'mapa_matriz_sin_id.jpeg')

    def test_dibujar_forma_nombrada(símismo):
        frm = mapa.FormaDinámicaNombrada(arch_frm_nombrada, col_id='COD_MUNI')
        frm.estab_valores({id_: random.random() for id_ in frm.ids})
        símismo._verificar_dibujó(frm, 'mapa_dict.jpeg')

    def test_dibujar_escala_valores(símismo):
        frm = mapa.FormaDinámicaNumérica(arch_frm_numérica, col_id='Id')
        frm.estab_valores(np.random.rand(*frm.valores.shape), escala_valores=(0, 2))
        símismo._verificar_dibujó(frm, 'mapa_escala.jpeg')

    def test_dibujar_unidades(símismo):
        frm = mapa.FormaDinámicaNumérica(arch_frm_numérica, col_id='Id')
        frm.estab_valores(np.random.rand(*frm.valores.shape), unidades='Cosos')
        símismo._verificar_dibujó(frm, 'mapa_unids.jpeg')

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.dir_)


class TestMapaResultados(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dir_ = tempfile.mkdtemp(prefix='prb_mapa_res_')

    def test_mapa_de_simul(símismo):
        frm = mapa.FormaDinámicaNumérica(arch_frm_numérica, col_id='Id')
        extern = {'Vacío': np.arange(len(frm.ids))}
        res = ModeloPrueba(dims=(215,)).simular(t=10, extern=extern)
        mapa.dibujar_mapa_de_res(forma_dinámica=frm, res=res, var='Vacío', t=3, directorio=símismo.dir_)

    def test_mapa_de_simul_grupo(símismo):
        ops = OpsSimulGrupo(t=3, extern=[{'Vacío': 1}, {'Vacío': 3}], nombre=['701', '101'])
        res = ModeloPrueba().simular_grupo(ops)

        frm = mapa.FormaDinámicaNombrada(arch_frm_nombrada, col_id='COD_MUNI')
        mapa.dibujar_mapa_de_res(forma_dinámica=frm, res=res, var='Vacío', t=3, directorio=símismo.dir_)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.dir_)



import re

import numpy as np


def leer_arch_egr(archivo, años=None, procesar=True):
    años = [años] if isinstance(años, int) else años

    dic_datos = {}

    n_est, n_polí = _buscar_forma(archivo)
    n_años = _buscar_n_años(archivo) if años is None else len(años)
    with open(archivo, 'r') as d:

        for a in _avanzar_año(d, año=años):

            for e in range(1, n_est + 1):
                _avanzar_a_estación(e, d)

                for p in range(1, n_polí + 1):
                    _avanzar_a_polígono(p, d)

                    f = d.readline()

                    while not _fin_sección(f):
                        vrs = _extraer_vars(f)
                        for var, val in vrs:
                            if var not in dic_datos:
                                dic_datos[var] = np.full((n_años, n_est, n_polí), np.nan)
                            if len(val):
                                dic_datos[var][a, e - 1, p - 1] = val.replace(',', '.')

                        f = d.readline()
    if procesar:
        procesar_cr(dic_datos)

    return dic_datos


def procesar_cr(dic):
    # Estos variables, si faltan en el egreso, deberían ser 0 y no NaN
    for cr in ['CrA', 'CrB', 'CrU', 'Cr4', 'A', 'B', 'U']:
        dic[cr][np.isnan(dic[cr])] = 0

    # Ajustar la salinidad por la presencia de varios cultivos
    kr = dic['Kr']

    salin_suelo = np.zeros_like(kr)

    # Crear una máscara boleana para cada valor potencial de Kr y llenarlo con la salinidad correspondiente
    kr0 = (kr == 0)
    salin_suelo[kr0] = \
        dic['A'][kr0] * dic['CrA'][kr0] + dic['B'][kr0] * dic['CrB'][kr0] + dic['U'][kr0] * dic['CrU'][kr0]

    kr1 = (kr == 1)
    salin_suelo[kr1] = dic['CrU'][kr1] * dic['U'][kr1] + dic['C1*'][kr1] * (1 - dic['U'][kr1])

    kr2 = (kr == 2)
    salin_suelo[kr2] = dic['CrA'][kr2] * dic['A'][kr2] + dic['C2*'][kr2] * (1 - dic['A'][kr2])

    kr3 = (kr == 3)
    salin_suelo[kr3] = dic['CrB'][kr3] * dic['B'][kr3] + dic['C3*'][kr3] * (1 - dic['B'][kr3])

    kr4 = (kr == 4)
    salin_suelo[kr4] = dic['Cr4'][kr4]

    para_llenar = [
        {'másc': kr0, 'cr': ['Cr4']},
        {'másc': kr1, 'cr': ['CrA', 'CrB', 'Cr4']},
        {'másc': kr2, 'cr': ['CrB', 'CrU', 'Cr4']},
        {'másc': kr3, 'cr': ['CrA', 'CrU', 'Cr4']},
        {'másc': kr4, 'cr': ['CrA', 'CrB', 'CrU']}
    ]

    for d in para_llenar:
        l_cr = d['cr']
        másc = d['másc']

        for cr in l_cr:
            dic[cr][másc] = salin_suelo[másc]

    # Aseguarse que no quedamos con áreas que faltan
    for k in ["A", "B"]:
        dic[k][dic[k] == -1] = 0


def _buscar_forma(arch):
    n_est = n_polí = 0
    # para hacer: más elegante
    with open(arch, 'r') as d:
        f = d.readline()
        while not re.match(r'\W*YEAR:\W+0+\W', f):
            f = d.readline()

        while not re.match(r'\W*YEAR:\W+1+\W', f):
            m = re.search(r'\W*Season:\W+([0-9]+)\W', f)
            if m:
                n_est = int(m.groups()[0])
            else:
                m = re.search(r'\W*Polygon:\W+([0-9]+)\W', f)
                if m:
                    n_polí = int(m.groups()[0])
            f = d.readline()

    return n_est, n_polí


def _buscar_n_años(arch):
    n_años = 0
    with open(arch, 'r') as d:
        for f in d:
            m = re.search(r'\W*Polygon:\W+([0-9]+)\W', f)
            if m:
                n_años = int(m.groups()[0]) + 1

    return n_años


def _avanzar_año(d, año=None):
    f = d.readline()
    if año is None:
        while re.match(r'\W*YEAR:\W+[0-9]+\W', f) is None:
            f = d.readline()
        yield int(re.search(r'\W*YEAR:\W+([0-9]+)', f).groups()[0])
    else:
        for a in año:
            while re.match(r'\W*YEAR:\W+%i\W' % a, f) is None:
                f = d.readline()
            yield a


def _avanzar_a_estación(estación, d):
    f = d.readline()
    while re.match(r'\W*Season:\W+%i\W' % estación, f) is None:
        f = d.readline()


def _avanzar_a_polígono(polígono, d):
    f = d.readline()
    while re.match(r'\W*Polygon:\W+%i\W' % polígono, f) is None:
        f = d.readline()


def _fin_sección(f):
    return re.match(r' #$', f) is not None


def _extraer_vars(f):
    return re.findall(r'([A-Za-z0-9*#]+)\W+=\W+([0-9.]+(?:E?[+\-0-9]+)?)?', f)
